# CaeFemMain -- 2022-09-27
# Copyright (C) [2022] [Ulrich Haeussler-Combe]
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License (GNU GPLv3) as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program; if not, see <http://www.gnu.org/licenses
#
from numpy import dot, array, mean, sqrt, arcsin, zeros
from math import cos, sin, pi, fabs
#from matplotlib.pyplot import figure, plot, grid, title, text, contour, clabel, show, axis, xticks, yticks, ylim, annotate
#from numpy.linalg import norm, det
from ConFemElem import ZeroD, C3D8, CPE4
from ConFemInOut import ReadOptionsFile #, PlotNodes
import matplotlib as mpl
import matplotlib.pyplot as plt
from os import path as pth
import numpy as np
import matplotlib.colors as colors
import mpl_toolkits.mplot3d as a3
import pylab as pl
import time as ti
import sys
import os
sys.path.insert(1, '../src')
#import imp
from ConFemBasics import SamplePoints

Colors = ['blue','black','darkviolet','magenta','indianred','darkred','cyan','yellow','blue']
Colors = ['tab:green','tab:brown','magenta','darkviolet','blue','green','yellow']
Colors = ['royalblue','darkviolet','green','magenta','blue','darkcyan','yellow']
Colors2 = ['red','green','magenta','cyan']
ColorTable = {}
#ColorTable['IsoD'] = 'white'
ColorTable['ELASTIC'] = 'black'
ColorTable['BOND'] = 'green'
ColorTable['ELASTICLT'] = 'gray'
ColorTable['ElasticSDA'] = 'black'
ColorTable['MISES'] = 'red'
ColorTable['MICRODAMAGE'] = 'magenta'
ColorTable['Lubliner'] = 'cyan'
ColorTable['ISODAMAGE'] = 'cyan'
ColorTable['ELASTICLT_SDA'] = 'magenta'
ColorTable['RESHEET'] = 'red'
ColorTable['NLSLAB'] = 'red'

#Colors = ['tab:green','tab:brown','magenta','darkviolet','blue','green','yellow']
#Colors = ['royalblue','magenta','darkviolet','darkcyan','blue','green','yellow']
FonSizTi='x-large'   # fontsize title x-large
FonSizAx='x-large'     # fontsize axis
#FonSizAx='medium'     # fontsize axis
LiWiCrv=3            # line width curve in pt
LiWiCrvMid=2            # line width curve in pt
LiWiCrvThin=1.0        # line width curve in pt


def DefPlot(Label):
    P0 = plt.figure()
    p0 = P0.add_subplot(111)
    p0.set_title(Label,fontsize='x-large')
    p0.tick_params(axis='x', labelsize='large')
    p0.tick_params(axis='y', labelsize='large')
    p0.grid()
    return P0, p0 

# from Ahmad 2020
def vtkPrincipal( NoList, ElemList, Name, timeStr):                         # from element Data   
    def Princstresses3D(sigma_x, sigma_y, sigma_z, tau_yz, tau_xz, tau_xy):
        
        AA = np.array([[sigma_x, tau_xy, tau_xz],
                       [tau_xy, sigma_y, tau_yz],
                       [tau_xz, tau_yz, sigma_z]])
            
        eigenvalue, eigenvector = np.linalg.eig(AA)
        eigenvalue_list = [eigenvalue[0], eigenvalue[1], eigenvalue[2]] 
        ma, mi = eigenvalue_list.index(max(eigenvalue)), eigenvalue_list.index(min(eigenvalue)) 
        v1= np.array(np.multiply(eigenvalue[ma], eigenvector[:,ma]))
        v3= np.array(np.multiply(eigenvalue[mi], eigenvector[:,mi]))
        for i in range(3):
            if i != ma and i != mi:
                v2= np.array(np.multiply(eigenvalue[i], eigenvector[:,i]))
                sig_2 = eigenvalue[i]
        sig_1 , sig_3 = eigenvalue[ma], eigenvalue[mi] 
        return sig_1, sig_2, sig_3, v1, v2, v3
    def Princstresses2D( xx, yy, xy ):                # calculation of principal stresses and vectors in 2D 
        delta = sqrt((-xx-yy)**2 - 4*(-xy**2 + xx*yy))
        b = -xx-yy
        s1 = (-b + delta)/2
        s2 = (-b - delta)/2
        
        if abs(s1 - xx) < ZeroD:                                # if first principle equals to sig xx
            vctr_1 = [1. , 0., 0.]
            vctr_2 = [0. , 1., 0.]
        elif abs(s1 - yy) < ZeroD:                              # if first principle equals to sig yy
            vctr_1 = [0. , 1., 0.]
            vctr_2 = [1. , 0., 0.]
        else:
            a1 = (s1-xx)/xy
            l1 = sqrt(1**2 + a1**2)
            vctr_1 = [1/l1 , a1/l1, 0.]
                
            a2 = (s2-xx)/xy
            l2 = sqrt(1**2 + a2**2)
            vctr_2 = [1/l2 , a2/l2, 0.]
        #if s1 < 0: s1 = 0.                  # only tension 
        #if s2 > 0: s2 = 0.                  # only compression
        return s1, [s1*vctr_1[0],s1*vctr_1[1],s1*vctr_1[2]], s2, [s2*vctr_2[0],s2*vctr_2[1],s2*vctr_2[2]]
    def ElemIntPointCoordinates( Elem, NoList):                             # integration point coordinates of element
        xyP, z, xy = [], [], []
        for i in Elem.Inzi:                                                 # ordered (without CM etc) nodal coordinates 
            node = NoList[i]
            if   Elem.SDim==1: xy += [node.XCo]
            elif Elem.SDim==2: xy += [node.XCo,node.YCo]
            elif Elem.SDim==3: xy += [node.XCo,node.YCo,node.ZCo]
        for j in range(Elem.nIntL):                                         # loop over integration points of element
            rst = SamplePoints[Elem.IntT,Elem.nInt-1,j]                     # local integration point coordinates - SamplePoints different to SamplePoints_
            XX  = Elem.FormX_( rst[0], rst[1], rst[2] )
            xyP = dot(XX,xy)                                                # global coordinates of integration point XXX
            z.append(xyP.tolist())
        return z                  
    # collect data before writing to vtu
    class Inptdata():
        def __init__(self, coord, intpdisp, sigma, vctr):
            self.coord    = coord
            self.intpdisp = intpdisp
            self.sigma    = sigma
            self.vctr     = vctr
    def intpoint( NoList, ElemList):
        intP = []
        for i in ElemList:
            if i.Type in ["CPS3", "CPS3_SDA", "CPS4", "CPS4_SDA", "C3D8", "C3D8S"]:
                noDisp = []
                for j in i.Inzi:
                    for k in range(len(NoList[j].BufDis)): 
                        noDisp.append(NoList[j].BufDis[k])                  # collect dofs of element from data read in
                IntpGlobal = ElemIntPointCoordinates( i, NoList)            # integration point coordinates
                for j in range(i.nIntL): 
                    Sig  = i.Data[j].tolist()                               # sigma from element-Data    
                    Intp = SamplePoints[i.IntT,i.nInt-1,j]
                    intpdisp = dot(i.FormN(Intp[0], Intp[1], Intp[2]), noDisp)
                    if i.Type in ["CPS4", "CPS4_SDA","CPS3", "CPS3_SDA"]:   
#                        if i.ElSet == "Supp": sig_1, v1, sig_2, v2 = 0., [0., 0., 0.], 0., [0., 0., 0.]
#                        else : 
                        sig_1, v1, sig_2, v2 = Princstresses2D(float(Sig[4]), float(Sig[5]), float(Sig[7]))
                        intP += [Inptdata([float(IntpGlobal[j][0]), float(IntpGlobal[j][1]), 0.], [ intpdisp[0], intpdisp[1], 0. ], [sig_1, sig_2, 0],[v1,v2, [0., 0., 0.]])]
                    elif i.Type in ["C3D8", "C3D8S"]: 
#                        if i.Set in ["Supp", "Supp1"]: sig_1, v1, sig_2, v2, sig_3, v3 = 0., [0., 0., 0.], 0., [0., 0., 0.], 0., [0., 0., 0.] 
#                        else:                          
                        sig_1,sig_2,sig_3,v1,v2,v3 = Princstresses3D(float(Sig[0]), float(Sig[1]), float(Sig[2]), float(Sig[3]), float(Sig[4]), float(Sig[5]))
                        intP += [Inptdata([float(IntpGlobal[j][0]), float(IntpGlobal[j][1]), float(IntpGlobal[j][2])], [ intpdisp[0], intpdisp[1], intpdisp[2] ], [sig_1, sig_2, sig_3],[v1,v2,v3])] 
        return intP  
    def WriteprincFile(fileOut_princ):
        f1 = open(fileOut_princ, 'w')
        f1.write('<VTKFile type="UnstructuredGrid" version="0.1">\n')
        f1.write('  <UnstructuredGrid>\n')
        f1.write('    <Piece NumberOfPoints="' + str(len(intP)) + '" NumberOfCells="' + str(len(intP)) + '">\n')
        #
        f1.write('      <Points>\n')
        f1.write('        <DataArray type="Float64" NumberOfComponents="3" format="ascii"/>\n')
        for i in intP:             f1.write('%12.4e%12.4e%12.4e\n' % (i.coord[0], i.coord[1], i.coord[2]))
        f1.write('      </Points>\n')
        f1.write('      <Cells>\n')
        f1.write('        <DataArray type="Int32" Name="connectivity" format="ascii"/>\n')
        for i in range(len(intP)): f1.write('       %3i\n' % (i))
        f1.write('        <DataArray type="Int32" Name="offsets" format="ascii"/>\n')
        for i in range(len(intP)): f1.write('       %3i\n' % (i))
        f1.write('        <DataArray type="Int32" Name="types" format="ascii"/>\n')
        for i in range(len(intP)): f1.write('       1\n')
        f1.write('      </Cells>\n')
        #
        f1.write('      <PointData>\n')
        f1.write('        <DataArray type="Float64" Name="IntP_displ" NumberOfComponents=" 3" format="ascii"/>\n')
        for i in intP:             f1.write('%12.4e%12.4e%12.4e\n' % (i.intpdisp[0], i.intpdisp[1], i.intpdisp[2])) 
        f1.write('        <DataArray type="Float64" Name="Vector_1" NumberOfComponents=" 3" format="ascii"/>\n')
        for i in intP:             f1.write('%12.4e%12.4e%12.4e\n' % (i.vctr[0][0], i.vctr[0][1], i.vctr[0][2]))
        f1.write('        <DataArray type="Float64" Name="Vector_2" NumberOfComponents=" 3" format="ascii"/>\n')
        for i in intP:             f1.write('%12.4e%12.4e%12.4e\n' % (i.vctr[1][0], i.vctr[1][1], i.vctr[1][2]))
        f1.write('        <DataArray type="Float64" Name="Vector_3" NumberOfComponents=" 3" format="ascii"/>\n')
        for i in intP:             f1.write('%12.4e%12.4e%12.4e\n' % (i.vctr[2][0], i.vctr[2][1], i.vctr[2][2]))  
        f1.write('        <DataArray type="Float64" Name="Princpal_Stresses" NumberOfComponents=" 3" format="ascii"/>\n')
        for i in intP:             f1.write('%12.4e%12.4e%12.4e\n' % (i.sigma[0], i.sigma[1], i.sigma[2])) 
        f1.write('      </PointData>\n')
        f1.write('    </Piece>\n')
        f1.write('  </UnstructuredGrid>\n')
        f1.write('</VTKFile>\n')
        f1.close()          
    # main part
    intP = intpoint( NoList, ElemList)                                      # collect data
    FFF = Name+'_Principal_ParaView'
    if pth.exists(FFF): pass
    else: os.mkdir(FFF)
    fileOut_princ = FFF + '/Principal_'+timeStr+'.vtu'
    WriteprincFile(fileOut_princ)                                           # write data     

# from Ahmad 2020
def vtkCrack(ElemList, fileName, timeStr):
        CracNodes, Tr, WL = [], [], []
        ne = 0
        Flag3D = False
        for i in ElemList:
                tr, wl = 0, 0
                if i.Type in ["C3D8S"]:
                    ne += 1
                    Flag3D = True
                    for a in range(len(i.xyzC)):
                        tr += i.wwT[a][0]   # position of normal traction in 3D = [0]
                        wl += i.wwL[a][0]   
                    tr = tr/len(i.xyzC)
                    wl = wl/len(i.xyzC)
                    for b in i.xyzC:
                        CracNodes +=  [[b[0], b[1], b[2] ]]
                        Tr += [tr]              # same traction for all nodes of the crack is assumed - Ahmad
                        WL += [wl]
                elif i.Type in ["CPS4_SDA"]:
                    ne += 1
                    for a in range(len(i.xyC)):
                        tr += i.wwT[0][a][1]   # position of normal traction in 2D = [1]
                        wl += i.wwL[0][a][1]   
                    tr = tr/len(i.xyC)
                    wl = wl/len(i.xyC)
                    for b in i.xyC:
                        CracNodes +=  [[b[0], b[1], 0. ]]
                        Tr += [tr]              # same traction for all nodes of the crack is assumed - Ahmad
                        WL += [wl]
                        
        FFF = fileName+'_Cracks_ParaView'
        if pth.exists(FFF): pass
        else: os.mkdir(FFF)
        fN = FFF+ "/" + 'Cracks_'+timeStr+'.vtu'
        fp = open(fN,'w')
        fp.write('<VTKFile type="UnstructuredGrid" version="0.1">\n')
        fp.write('  <UnstructuredGrid>\n')
        fp.write('    <Piece NumberOfPoints='+'"'+str(len(CracNodes))+'"'+' NumberOfCells='+'"'+str(ne)+'">\n')
            #
        fp.write('      <Points>\n')
        fp.write('        <DataArray type="Float64" NumberOfComponents="3" format="ascii"/>\n')
        for i in CracNodes:
                fp.write("%12.4e%12.4e%12.4e\n"%(i[0],i[1],i[2]))
        fp.write('      </Points>\n')
            # cell connectivity, offset, types
        fp.write('      <Cells>\n')
        fp.write('        <DataArray type="Int32" Name="connectivity" format="ascii"/>\n')
        c = 0
        for i in range(ne):
            if Flag3D:
                fp.write('%8i%8i%8i%8i'%(0+c,1+c,2+c,3+c))         # only VTK_QUAD is implimented for crack surface !!! - Ahmad
                c += 4 
                fp.write('\n') 
            else:
                fp.write('%8i%8i'%(0+c,1+c))
                c += 2
                fp.write('\n')
        fp.write('        <DataArray type="Int32" Name="offsets" format="ascii"/>\n') 
        m = 1
        for i in range(ne):
            if Flag3D:
                fp.write('%8i'%(4*m))
                m += 1
                fp.write('\n')
            else:
                fp.write('%8i'%(2*m))
                m += 1
                fp.write('\n')
        fp.write('        <DataArray type="Int32" Name="types" format="ascii"/>\n')
        for i in range(ne):
            if Flag3D: fp.write('%8i\n'%(9))
            else:      fp.write('%8i\n'%(3))
        fp.write('      </Cells>\n')
            #
        fp.write('      <PointData>\n')
        fp.write('        <DataArray type="Float64" Name="Cra_traction" NumberOfComponents=" 1" format="ascii"/>\n')
        for i in range(len(CracNodes)):
                fp.write('%12.4e'%(Tr[i]))
                fp.write('\n')
        fp.write('        <DataArray type="Float64" Name="Cra_opening" NumberOfComponents=" 1" format="ascii"/>\n')
        for i in range(len(CracNodes)):
                fp.write('%12.4e'%(WL[i]))
                fp.write('\n')
        fp.write('      </PointData>\n')
            # 
        fp.write('    </Piece>\n')
        fp.write('  </UnstructuredGrid>\n')
        fp.write('</VTKFile>\n')
        fp.close()

def vtkFile(fileName, ElemList, VTKNodeList, marker, timeStr, elTypesExc, noTypesInc, offset, numSDA):
    # compute number of respective elements
    ne = 0
    for i in ElemList:
        if i.Type not in elTypesExc: ne += 1
    nn, nn_ = 0, 0
    for i in VTKNodeList:
        if i.Type in noTypesInc: nn += 1
    # write data
    FFF = fileName+'_'+marker+'_ParaView'
    if pth.exists(FFF): pass
    else: os.mkdir(FFF)
    fN = FFF+ "/" + marker+'_'+timeStr+'.vtu'
    fp = open(fN,'w')
    fp.write('<VTKFile type="UnstructuredGrid" version="0.1">\n')
    fp.write('  <UnstructuredGrid>\n')
    fp.write('    <Piece NumberOfPoints='+'"'+str(nn)+'"'+' NumberOfCells='+'"'+str(ne-numSDA)+'">\n')
    #
    # write node coordinates to vtk
    #
    fp.write('      <Points>\n')
    fp.write('        <DataArray type="Float64" NumberOfComponents="3" format="ascii"/>\n')
    for i in VTKNodeList:
        if i.Type in noTypesInc: fp.write("%12.4e%12.4e%12.4e\n"%(i.XCo,i.YCo,i.ZCo))
    fp.write('      </Points>\n')
    #
    # cell connectivity, offset, types to vtk
    #
    fp.write('      <Cells>\n')
    fp.write('        <DataArray type="Int32" Name="connectivity" format="ascii"/>\n')
    for i in ElemList:
        if i.Active and not i.Type in elTypesExc:
            for j in i.Inzi: fp.write('%8i'%(j-offset))    # this requires collection of nodes belonging to a common type of elements in input file !!! 
            fp.write('\n')
    fp.write('        <DataArray type="Int32" Name="offsets" format="ascii"/>\n') 
    off = 0
    for i in ElemList:
        if i.Active and not i.Type in elTypesExc:
            fp.write('%8i\n'%(len(i.Inzi)+off))
            off += len(i.Inzi)
    fp.write('        <DataArray type="Int32" Name="types" format="ascii"/>\n')
    for i in ElemList:
        if i.Active and not i.Type in elTypesExc:
            if   i.Type in ["C3D8","C3D8_EFG", "C3D8S"]: type_ = 12
            elif i.Type in ["C3D8_IGA"]:                    type_ = 25 
            elif i.Type in ["C3D4"]:                        type_ = 10
            elif i.Type in ["T3D3E"]:                       type_ = 21
            elif i.Type in ["T3D2E", "T2D2E", "T2D2", "B23I", "B23"]:    type_ = 3
            elif i.Type in ["CPS4", "CPS4_SDA"]:            type_ = 9
            elif i.Type in ["CPS3", "CPS3_SDA"]:            type_ = 5
            else:  raise NameError("CaeFemInOut::DataOutvtk: unknown element type 2",i.Type)
            fp.write('%8i\n'%(type_))
    fp.write('      </Cells>\n')
    #
    # point/node meta data
    #
    fp.write('      <PointData>\n')
    stressName = noTypesInc[0]+'_stress'
    if len(noTypesInc)>1: raise NameError("ConFemInOut::DataOutVTK: 1")
    if   noTypesInc[0] in ["T2DE", "T3DE"]: num=1
    elif noTypesInc[0] in ["T2DE", "T3DE"]: num=2  # ???????????????????????????????????????
    else:                                   num=6
    #
    # write node stresses to vtk, BufSig comes from first part of PostElemVTK
    #
    fp.write('        <DataArray type="Float64" Name="'+stressName+'" NumberOfComponents="'+str(num)+'" format="ascii"/>\n')
    BondFlag = False
    mattype = False
    for i in VTKNodeList:
        if i.Type in noTypesInc:
            if i.Type in ["T2DE", "T3DE"]:                                  # Type determined in EvNodeEl
                fp.write('%12.4e'%(i.BufSig[1]))
                BondFlag = True
            elif i.Type in ["B23I"]:
                fp.write('%12.4e%12.4e'%(i.BufSig[0], i.BufSig[1]))
                BondFlag = True
            elif i.Type in ["C2D"]:
                for j in i.BufSig: fp.write('%12.4e'%(j))
                mattype = True
            else:
                for j in i.BufSig: fp.write('%12.4e'%(j))
            fp.write('\n')
    #
    # write material to vtk
    #        
    if mattype:                                                             # for flag mattype see above
        MatNamee = noTypesInc[0]+'_Mat'
        fp.write('        <DataArray type="Float64" Name="'+MatNamee+'" NumberOfComponents="1" format="ascii"/>\n')
        for i in VTKNodeList:
            if i.Type in noTypesInc:
#                if i.material == "IsoDam": fp.write('%12.4e'%(i.BufSig[3])) # currently for C2D4/CPE4
#                else:                      fp.write('%12.4e'%(1))
                fp.write('%12.4e'%(i.BufSig[3]))                            # currently for C2D4/CPE4
                fp.write('\n')
    #
    #
    #
    if BondFlag:
        bondName = noTypesInc[0]+'_bond'
        fp.write('        <DataArray type="Float64" Name="'+bondName+'" NumberOfComponents="1" format="ascii"/>\n')
        for i in VTKNodeList:
            if i.Type in noTypesInc:
#                fp.write('%12.4e'%(i.BufBond))
                fp.write('%12.4e'%(i.BufSig[5]))
                fp.write('\n')
        
        slipName= noTypesInc[0]+'_slip'
        fp.write('        <DataArray type="Float64" Name="'+slipName+'" NumberOfComponents="1" format="ascii"/>\n')
        for i in VTKNodeList:
            if i.Type in noTypesInc:
#                fp.write('%12.4e'%(i.Bufslip))
                fp.write('%12.4e'%(i.BufSig[2]))
#                    if i.Label == 641:  fslip.write('%12.4e\n'%(i.Bufslip))
                fp.write('\n')
#            fslip.flush()
    #
    # write displacements to vtk
    #    
    displName = noTypesInc[0]+'_displ'
    fp.write('        <DataArray type="Float64" Name="'+displName+'" NumberOfComponents=" 3" format="ascii"/>\n')
    for i in VTKNodeList:
        if i.Type in noTypesInc:
            if i.Type in ["B23I"]: fp.write("%12.4e%12.4e%12.4e"%(i.BufDis[0],i.BufDis[1],0))
            else:
                for j in i.BufDis: fp.write('%12.4e'%(j))
                if len(i.BufDis)==2: fp.write('%12.4e'%(0.))
            fp.write('\n')
    fp.write('      </PointData>\n')
    # 
    fp.write('    </Piece>\n')
    fp.write('  </UnstructuredGrid>\n')
    fp.write('</VTKFile>\n')
    fp.close()
    return nn
#    return nn_

def PostElemVTK( FileDir, FileName, timeStr, ElList, NodeList,NoIndToCMInd, VecU, ElemResults,NodeResults, LinAlgFlag, Flag3D):
    def Distribute():
        for k,i in enumerate(el.Inzi):                                      # loop over element nodes
            no = NodeList[NoIndToCMInd[i]]
            for j in range(DatLen): no.BufSig[j] += X[j][k]                 # values of same type in "row" of X with "column" for node
            no.BufCount += 1
    # end def    
    for i in NodeList:                                                      # initialize and fill displacement buffer
        i.BufSig = zeros((6),dtype=float)                                   # buffer for nodal stress values used for VTK - be cautious with its length, must cover all elements
#        i.BufBond, i.Bufslip = 0., 0.
#        i.BufBondCount = 0
        i.BufCount = 0
        i.material = ""
        key = i.Label
        try:    i.BufDis = NodeResults[key][2]                              # displacement buffer
        except: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 5")
    #
    numSDA = 0                                                              # total number of SDA elements - Ahmad
    for el in ElList:
        if el.Active:
            IntLen = el.nIntL                                               # number of integration points of element
            #
            # ElemResults comes from  ReadResultsFromFile, see below
            #
            if el.Type in ["C3D8","C3D8S"]:
                if IntLen != 8: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 3")
                DatLen = 6                                                  # corresponds to value types - actually 6 stress items
                Buf, X = zeros((DatLen,IntLen), dtype=float), []
                for i in range(IntLen):
                    key = str(el.Label)+str(el.Set)+str(i)
                    try:    ElRes = ElemResults[key][5][3:9]
                    except: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 1")
                    for j in range(DatLen): Buf[j,i] += [ElRes[j]]          # reorder rows of ElRes in columns of Buf so that rows of Buf hold values of same type
                for i in range(DatLen):                                     # loop over value types
                    X += [dot(C3D8.FormNI(el),Buf[i])]                      # multiply Buf-row with inverse of shape function to derive nodal values for each value type in row of
                if el.Type in ["C3D8S"]: numSDA += 1
                Distribute()
            elif el.Type in ["CPS4","CPE4","CPS4S","CPE4S"]:
                if IntLen != 4: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 8")
                DatLen = 4                                                  # corresponds to value types - actually 6 stress items
                Buf, X = zeros((DatLen,IntLen), dtype=float), []
                for i in range(IntLen):
                    key = str(el.Label)+str(el.Set)+str(i)
                    try:
                        ElRes_ = ElemResults[key][5]
                        if len(ElRes_)==11: ElRes = array([ElemResults[key][5][6],ElemResults[key][5][7],ElemResults[key][5][9],ElemResults[key][5][10]]) # last for damage of IsoDam
                        else:               ElRes = array([ElemResults[key][5][6],ElemResults[key][5][7],ElemResults[key][5][9],0.])
                    except: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 9")
                    for j in range(DatLen): Buf[j,i] += [ElRes[j]]          # reorder rows of ElRes in columns of Buf so that rows of Buf hold values of same type
                for i in range(DatLen):                                     # loop over value types
                    X += [dot(CPE4.FormNI(el),Buf[i])]                      # multiply Buf-row with inverse of shape function to derive nodal values for each value type in row of
                if el.Type in ["CPS4S","CPE4S"]: numSDA += 1
                Distribute()
            elif el.Type in ["T3D2E"]:
                if IntLen != 2: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 5")
                DatLen = 6                                                  # corresponds to value types - actually strain, stress for T3D3E, 6 bond items: bond item 1 for long slip, item 4 for long stress 
                ElRes, X = [], zeros((DatLen,2), dtype=float)               # 2 for three nodes of T3D2E
                for i in range(IntLen):
                    key = str(el.Label)+str(el.Set)+str(i)                  # key is label, set, integration point
                    try:    ElRes += [ElemResults[key][5][3:9]]
                    except: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 6")
                for i in range(DatLen):                                     # loop over value types
                    f1 = ElRes[0][i]                                        # first integration point value 
                    f2 = ElRes[1][i]                                        # second integration point value
                    X[i,0] =  +1.366025404*f1 - .3660254039*f2              # extrapolation to left node, see ExtrapolateIntpointsToNodesT3D3E_mws.pdf
                    X[i,1] =   -.366025404*f1 + 1.366025404*f2              # extrapolation to right node
                Distribute()
            elif el.Type in ["T3D3E"]:
                # special treatment to determine global displacement of center node
                no0 = NodeList[ NoIndToCMInd[el.Inzi[0]] ] 
                no1 = NodeList[ NoIndToCMInd[el.Inzi[1]] ] 
                no2 = NodeList[ NoIndToCMInd[el.Inzi[2]] ] 
                uEl = array([ no0.BufDis[0],no0.BufDis[1],no0.BufDis[2] , no1.BufDis[0],no1.BufDis[1],no1.BufDis[2], no2.BufDis[0]])
                NN = el.FormN( 0., 0., 0. )
                no2.BufDis = array([0.,0.,0.])                              # overwrite initial value
                for i, u in enumerate(dot(NN,uEl)): no2.BufDis[i] = u
                #
                if IntLen != 2: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 2")
                DatLen = 6                                                  # corresponds to value types - actually strain, stress for T3D3E, 6 bond items: bond item 1 for long slip, item 4 for long stress 
                ElRes, X = [], zeros((DatLen,3), dtype=float)               # 3 for three nodes of T3D3E
                for i in range(IntLen):
                    key = str(el.Label)+str(el.Set)+str(i)                  # key is label, set, integration point
                    try:    ElRes += [ElemResults[key][5][3:9]]
                    except: raise NameError("ConFemPostProcStandAalone::PostElemVTK: 4")
                for i in range(DatLen):                                     # loop over value types
                    f1 = ElRes[0][i]                                        # first integration point value 
                    f2 = ElRes[1][i]                                        # second integration point value
                    X[i,0] =  +1.366025404*f1 - .3660254039*f2              # extrapolation to left node, see ExtrapolateIntpointsToNodesT3D3E_mws.pdf
                    X[i,1] =   -.366025404*f1 + 1.366025404*f2              # extrapolation to right node
                    X[i,2] =   .4999999998*f1 + .5000000002*f2              # interpolation to center node
                Distribute()
    #
    for no in NodeList:
        if no.BufCount>0:
            no.BufSig = no.BufSig/no.BufCount

#        print(f'{timeStr:s}, {no.Label:4d}, {no.Type:5s}, {", ".join(z for z in [f"{x:8.4f}" for x in no.BufSig[:]])}')
    # control
#    for el in ElList:
#        if el.Active and el.Type in ["T3D3E"]:
#            for i in el.Inzi:                       
#                no = NodeList[NoIndToCMInd[i]]
#                print(el.Label,no.Label,no.XCo,no.YCo,no.ZCo,no.BufCount,'__',no.BufSig[0],no.BufSig[1])
 
    off = 0
    ContinuumFirst = True
    fileName=FileDir+FileName
    if LinAlgFlag: NodeList.sort(key=lambda t: t.Index)
    if ContinuumFirst:
        if Flag3D:
            #     vtkFile(fileName, NodeList, VTKNodeList,  marker,timeStr,   elTypesExc,                                        noTypesInc, offset, numSDA)
            off = vtkFile(fileName, ElList,NodeList, "C3D",  timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","T3D3E","T3D2E"], ["C3D"], off, numSDA)
            off = vtkFile(fileName, ElList,NodeList,    "T3D",  timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","C3D4","C3D8","C3D8_EFG","C3D8_IGA","C3D8S"],  ["T3DE"], off, 0) # take care of collection of nodes belonging to a common type of elements in input file !!!
        else:
            off = vtkFile(fileName, ElList,NodeList, "CPS4", timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","T2D3E","T2D2E","T2D2", "B23I", "B23"], ["C2D"], off, numSDA)
            if False: #Beam:
                off = vtkFile(fileName, ElList,NodeList,    "B23",  timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","CPS4","CPS4_EFG","CPS4_IGA","CPS4_SDA","CPS3","CPS3_SDA"],  ["B23I"], off, 0) 
            else:
                off = vtkFile(fileName, ElList,NodeList,    "T2D",  timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","CPS4","CPS4_EFG","CPS4_IGA","CPS4_SDA","CPS3","CPS3_SDA"],  ["T2DE"], off, 0) # take care of collection of nodes belonging to a common type of elements in input file !!!
        vtkCrack(ElList, fileName, timeStr)
        vtkPrincipal( NodeList, ElList, fileName, timeStr)
    else:
        off = vtkFile(fileName, ElList,NodeList,    "T3D", timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","C3D8","C3D8_EFG","C3D8_IGA","C3D8S"],  ["T3DE"], off, 0) # take care of collection of nodes belonging to a common type of elements in input file !!!
        off = vtkFile(fileName, ElList,NodeList, "C3D", timeStr, ["B2D2E","B2D3E","B3D2E","B3D3E","T3D3E","T3D2E"], ["C3D"], off, numSDA)
    if LinAlgFlag: NodeList.sort(key=lambda t: t.CMIndex)

def PostElem1D( f2, ResultTypes, PlotTimes ):
    def Rotate( dataXe, dataYe):
        n = len(dataXe)
        if n>1:
            LL = ((dataXe[n-1]-dataXe[0])**2+(dataYe[n-1]-dataYe[0])**2)**0.5
            cosA = (dataXe[n-1]-dataXe[0])/LL
            sinA = (dataYe[n-1]-dataYe[0])/LL
            return cosA, sinA
        else: return 1, 0
    dataXe = []                                                     # list for ip x per element
    dataYe = []                                                     # list for ip y per element
    dataAe = []                                                     # list for ip data per element
    dataXt = []                                                     # list for x per time step
    dataYt = []                                                     # list for y per time step
    dataAt = []                                                     # list for A per time step
    dataSt = []                                                     # list for sets per time step
    dataX  = []                                                     # list for all x
    dataY  = []                                                     # list for all y
    dataA  = []                                                     # list for all A
    dataS  = []                                                     # list for all sets
    Times  = []                                                     # list for time values
#    FlagData = False
    FlagElAll = False                                                       # to indicate whether at least one 1D element was hit
    # read all data from file
    z1 = f2.readline()
    z2 = z1.split()
    while z1!="":
        # new time
        if z2[0]=="Time":
            Time = float(z2[1])
            if Time in PlotTimes or len(PlotTimes)==0: PlotTFlag = True
            else:                                      PlotTFlag = False
            if PlotTFlag:
                ElSetI = 0
                ElSets = {}                                                 # dictionary: element set label -> index
                Times += [Time]
                if len(dataXe)>0:
                    dataXt += [dataXe]                                      # ip x coordinates: append last element of previous time
                    dataYt += [dataYe]                                      # ip y coordinates: append last element
                    dataAt += [dataAe]                                      # result data:   append last element of previous time
                    dataX  += [dataXt]                                      # append to list of all x
                    dataY  += [dataYt]                                      # append to list of all y
                    dataA  += [dataAt]                                      # append to list of all data
                    dataS  += [dataSt]                                      # 
                    dataXt, dataYt, dataAt, dataXe, dataYe, dataAe = [], [], [], [], [], [] # reset lists for elements
        # new element
        elif PlotTFlag and z2[0]=="El":
            ElType = z2[2]
            if ElType in ['CPS4','CPS4R','CPE4','CPE4R', 'CPS3','CPE3','CPS4S','CPE4S','CPS3S','CPE3S',
                          'SB3','SH4','SH3',
                          'T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E',
                          'B2D2E','B2D3E','B3D2E','B3D3E',
                          'C3D8','C3D8S']:
                FlagEl = False 
            else:
                FlagElAll = True
                FlagEl = True
                ElSet  = z2[3]
                if ElSet not in ElSets:                                     # new elset
                    ElSetI = ElSetI+1                                       # count elsets for later plot loop
                    ElSets[ElSet] = ElSetI                                  # index for new elset
                dataSt += [ElSetI]                                          # assign elset index to current element data
                if len(dataXe)>0:
                    dataXt += [dataXe]                                      # append to list of elements per time step
                    dataYt += [dataYe]                                      # append to list of elements per time step
                    dataAt += [dataAe]                                      # append to list of data
                    dataXe = []                                             # reset list for element
                    dataYe = []                                             # reset list for element
                    dataAe = []                                             # reset list for element
        # integration point data
        elif PlotTFlag and FlagEl:                                          # ip data line
            dataXe += [float(z2[0])]                                        # append ip x-coordinate to element list
            dataYe += [float(z2[1])]                                        # append ip y-coordinate to element list
            dataAe += [[float(z2[pos]) for pos in range(2,len(z2))]]        # append ip data to element list
        z1 = f2.readline()
        z2 = z1.split()
        # data reading finished
    #
    if FlagElAll: 
        dataXt += [dataXe]                                                  # append last element
        dataYt += [dataYe]                                                  # append last element
        dataAt += [dataAe]                                                  # append last element
        dataX  += [dataXt]                                                  # append to list of all
        dataY  += [dataYt]                                                  # append to list of all
        dataA  += [dataAt]                                                  # append to list of all
        dataS  += [dataSt]
#        if len(dataX)==0 or not FlagData: return 0
        if len(dataX)==0: return 0
    
        # plot whole stuff
        scal = [ 1., 1., 1., 1., 1., 1., 1., 1., 1., 1.]
    #    if raw_input('Separate element sets for plot? (y/n):')=='y':
        if len(ElSets)>0: SFlag=True
        else:             SFlag=False                                           # does not separate for elsets
        for i in range(len(Times)): print("PostElem1D process plot for time ", Times[i],Colors[i % len(Colors)]) # pre user info
        #
        for sI, ElS in zip( list(ElSets.values()) , list(ElSets.keys()) ):      # loop over element sets; value is elset index/counter, key is elset label 
            titleList = ResultTypes[ElS]                                        # ResultTypes set per elset in DataInput post
            for I, tL in enumerate(titleList):                                  # loop over result types
                Ymin, Ymax = 0,0
                if tL == None: continue
                P0 = plt.figure()
                p0 = P0.add_subplot(111)
                if SFlag: p0.set_title(ElS+': '+tL, fontsize=FonSizTi)          # ElS is elset label
                else:     p0.title(tL, fontsize=FonSizTi)
                p0.tick_params(axis='x', labelsize=FonSizAx)
                p0.tick_params(axis='y', labelsize=FonSizAx)
                p0.grid()
                for i in range(len(Times)):                                     # loop over time steps
                    Colr = Colors[i % len(Colors)]
                    X_, Y_ = [], []
                    dA, dX, dY, dS = dataA[i], dataX[i], dataY[i], dataS[i]     # element data, p x-coordinate, ip y-coordinate, elset-label for time i
                    ne = len(dA)                                                # number of elements
                    PlotThrough = False
                    for j in range(ne):                                         # loop over elements
                        if dS[j] == sI or not SFlag:                            # found element in current elset to plot  or   do not separate for elsets
                            neD = len(dA[j])                                    # number of data items per element
                            if neD==1: PlotThrough = True                       # only one data item per element
                            cosA, sinA = Rotate( dX[j], dY[j])                  # ip coordinates
    #                        X                   = [ dX[j][k] - sinA*(scal[I]*dA[j][k][I]) for k in range(neD)] # shifted origin x 
    #                        if PlotThrough: X_ += [ dX[j][0] - sinA*(scal[I]*dA[j][0][I]) ]
                            X                   = [ dX[j][k] + sinA*(scal[I]*dA[j][k][I]) for k in range(neD)] # shifted origin and and rotation applied to x = 0 
                            if PlotThrough: X_ += [ dX[j][0] + sinA*(scal[I]*dA[j][0][I]) ]
                            if fabs(sinA)>ZeroD:
                                p0.plot( dX[j], dY[j], '-k')                    # integration point coordinates  undeformed structure connected by line
                                Y =                  [ dY[j][k] + cosA*(scal[I]*dA[j][k][I]) for k in range(neD)]
                                if PlotThrough: Y_+= [ dY[j][0] + cosA*(scal[I]*dA[j][0][I])]
                            else: 
                                Y =  [                       (scal[I]*dataA[i][j][k][I]) for k in range(neD)]
                                if PlotThrough: Y_+= [       (scal[I]*dataA[i][j][0][I])]
                            Ymin=min(Ymin,min(Y))
                            Ymax=max(Ymax,max(Y))
                            p0.plot( X, Y, '-', color=Colr,linewidth=LiWiCrv)
                            p0.plot( X, Y, 'o', color='tab:red',ms=5.) 
                    if PlotThrough: p0.plot( X_, Y_, '-', color=Colr,linewidth=LiWiCrvThin)
    #                        print titleList[I], X, Y
    #                        plot( X, Y, '-', color=Colors[i % len(Colors)])
                p0.set_ylim(1.1*Ymin,1.1*Ymax)
                P0.autofmt_xdate()
    return 0

def PostScales( ElemList,SolSecs, ElemResults, ScaleStress):
    Sc = []                                                                 # collects scalings for all element sets
    ScDict = {}
    for S in SolSecs:
        ScaleElSet, PData = 1., 0.
        for sc in ScaleStress:
            if sc[0].upper()==S: 
                ScaleElSet = sc[1]                                          # superposed scaling factor for element set
                break 
#        if SolSecs[S].ElemTypes in [["S1D2"],["T1D2"],["B23"],["B23E"],["B23I"],["B23EI"],["B21"],["B21E"] ]: continue            # this routine is not appropriate for this element types
        if SolSecs[S].ElemTypes in [["S1D2"],["T1D2"],["B23"],["B23E"],["B21"],["B21E"] ]: continue            # this routine is not appropriate for this element types
        if   SolSecs[S].ElemTypes in [["B23I"],["B23EI"]]:  PData = 1
        elif SolSecs[S].ElemTypes in [["T2D2E"],["T2D3E"]]: PData = 2
        elif SolSecs[S].ElemTypes in [["SH4"],["SH3"]]:     PData = 3
        elif SolSecs[S].ElemTypes in [["SB3"]]:             PData = 4
        minSig, maxSig, minSig1, maxSig1, minSig2, maxSig2, meanCharL, nEl = 0.,0., 0.,0., 0.,0., 0., 0
        for e in SolSecs[S].Elems:                                         # loop over element indices of Solid Section
            i = ElemList[e]
            if i.Type in ["SH3","SH4"]: Lis = i.Lists1()
            else:                       Lis = range(i.nIntL) 
            meanCharL += (i.Lch_/sqrt(float(i.nIntL)))
            nEl       += 1
            # loop over integration points of element
            for j in Lis:
                key = str( i.Label)+i.Set+str(j)
                if key not in ElemResults: continue                         # e.g. for key with negative label, see PostElem2D stresses plotting
                Sig = DataToSig( ElemResults[key])
                if PData==1:                                                # for embedded beam elements
                    if min(Sig[0:3])<minSig:  minSig  = min(Sig[0:3])       # bending stresses
                    if max(Sig[0:3])>maxSig:  maxSig  = max(Sig[0:3])
                    if Sig[2]       <minSig1: minSig1 = Sig[2]              # bond stress
                    if Sig[2]       >maxSig1: maxSig1 = Sig[2]
                    if Sig[3]       <minSig2: minSig2 = Sig[3]              # lateral indentation stress
                    if Sig[3]       >maxSig2: maxSig2 = Sig[3]
                elif PData==2:                                              # for embedded truss elements
                    if Sig[0]       <minSig:  minSig  = Sig[0]              # longitudinal stress
                    if Sig[0]       >maxSig:  maxSig  = Sig[0]
                    if Sig[1]       <minSig1: minSig1 = Sig[1]              # bond stress
                    if Sig[1]       >maxSig1: maxSig1 = Sig[1]
                    if Sig[2]       <minSig2: minSig2 = Sig[2]              # lateral indentation stress
                    if Sig[2]       >maxSig2: maxSig2 = Sig[2]
                elif PData==3:                                              # for shell elements
                    if min(Sig[0:3])<minSig:  minSig  = min(Sig[0:3])       # membrane / Normal forces
                    if max(Sig[0:3])>maxSig:  maxSig  = max(Sig[0:3])
                    if min(Sig[3:6])<minSig1: minSig1 = min(Sig[3:6])       # bending moments
                    if max(Sig[3:6])>maxSig1: maxSig1 = max(Sig[3:6])
                elif PData==4:                                              # for slab elements
                    if min(Sig[0:3])<minSig:  minSig  = min(Sig[0:3])       # bending moments
                    if max(Sig[0:3])>maxSig:  maxSig  = max(Sig[0:3])
                    if min(Sig[3:5])<minSig1: minSig1 = min(Sig[3:5])       # shear forces
                    if max(Sig[3:5])>maxSig1: maxSig1 = max(Sig[3:5])
                else:                               
                    if min(Sig)<minSig: minSig = min(Sig)
                    if max(Sig)>maxSig: maxSig = max(Sig)
        meanCharL = meanCharL/nEl
        absSig  = max(maxSig,-minSig)
        absSig1 = max(maxSig1,-minSig1)
        absSig2 = max(maxSig2,-minSig2)
        sc, sc1, sc2 = 0., 0., 0.
#        if absSig<ZeroD: 
#            print("PostElem2D zero stress result section",S)
#            sc = [1.]
#        else:
        fac = ScaleElSet
        if PData in [1,2]:                                                  # beam, truss
            if absSig >ZeroD: sc  = nEl* meanCharL/absSig                                 # respective scaling factor
            if absSig1>ZeroD: sc1 = nEl* meanCharL/absSig1
            if absSig2>ZeroD: sc2 = nEl* meanCharL/absSig2
            if PData==1: sc  = [ fac*sc, fac*sc, fac*sc1, fac*sc2 ]
            else:        sc  = [ fac*sc, fac*sc1, fac*sc2 ]
        elif PData in [3,4]:                                                # shell, slab
            if absSig >ZeroD: sc  = meanCharL/absSig                                      # respective scaling factor
            if absSig1>ZeroD: sc1 = meanCharL/absSig1                                     # respective scaling factor
            sc  = [ fac*sc, fac*sc1 ]
        else:
            if absSig >ZeroD: sc = meanCharL/absSig                                       # respective scaling factor
            sc = [fac*sc]
        Sc += [sc]
        ScDict[S] = sc
    return Sc, ScDict
def DataToSig( ElemKey):                                                    # extracts integration point data from ElemResults data
    Data= ElemKey[5]
    # data specific for element type
    if   ElemKey[3] in ["CPS4","CPE4","CPS4S","CPE4S","CPS3","CPE3","CPS3S","CPE3S","CPS6","CPE6","CPS6S","CPE6S"]: 
        sig = [Data[6],Data[7],Data[9]]
    elif ElemKey[3] in ["T2D2","T2D3","T3D2","T3D3"]: 
        sig = [Data[3]]                                                     # Data should hold all floating point from elemout, [3] long. stress, [12] bond stress 
    elif ElemKey[3] in ["T2D2E","T2D3E"]: 
        sig = [Data[3],Data[6],Data[7]]                                   # Data should hold all floating point from elemout, [3] long. stress, [12] bond stress, [13] penalty stress 
    elif ElemKey[3] in ["B23I","B23EI"]: 
        sig = [Data[6],Data[7],Data[10],Data[11]]                           # low and up stress, long. and lat. bond stress
    elif ElemKey[3] in ["B23"]: 
        sig = [Data[6],Data[7]]                                             # low and up stress
    elif ElemKey[3] in ["SB3"]:
        sig = [Data[5],Data[6],Data[7],Data[8],Data[9]]                                     # bending moments
    elif ElemKey[3] in ["SH3","SH4"]:
        sig = [Data[3],Data[4],Data[5],Data[6],Data[7],Data[8]]             # membrane forces, bending moments
    elif ElemKey[3] in ["C3D8"]:
        sig = [Data[3],Data[4],Data[5],Data[6],Data[7],Data[8]]
    else: raise NameError("ConFem::PostElem2D: unknown element type for reading from files",ElemKey[3])
    return sig

def PostElem2D( ElemList,NodeList, ScaleU, SolSecs, MaPlLib, NoIndToCMInd, ElemResults,NodeResults, TimeEl, pdir, SCales ):
    def Annot( ):
        if Xmax!=None:
            p1.annotate(    'max s1 '+"%6.1e"%Smax+' s2 '+"%6.1e"%S2max+' '+'%6i'%(ELaMax),xy=(Xmax, Ymax),xycoords='data',xytext=(0.40,0.87),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='medium')
#            if ElType=='SH4': 
#                P1.annotate('max s1 '+'%7.4f'%(Smax_)+' s2 '+'%7.4f'%(S2max_),xy=(Xmax_, Ymax_),xycoords='data',xytext=(0.50,0.87),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='large')
        if Xmin!=None:
            p1.annotate(    'min s1 '+"%6.1e"%Smin+' s2 '+"%6.1e"%S2min+' '+'%6i'%(ELaMin),xy=(Xmin, Ymin),xycoords='data',xytext=(0.40,0.84),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='medium')
#            if ElType=='SH4': 
#                P1.annotate('min s1 '+'%7.4f'%(Smin_)+' s2 '+'%7.4f'%(S2min_),xy=(Xmin_, Ymin_),xycoords='data',xytext=(0.50,0.84),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='large')
    def PloPrin( pp, scaleP, XX, YY, p0, Smax, Smin, ElLabel, ELaMax_, ELaMin_, Xmax_, Ymax_, Xmin_, Ymin_):
        ELaMax, ELaMin, Xmax, Ymax, Xmin, Ymin = ELaMax_, ELaMin_, Xmax_, Ymax_, Xmin_, Ymin_
        Smax_ , Smin_, S2min, S2max = Smax, Smin, 0, 0
        SS = pp[0]                                                          # larger principal stress - signed
        if SS>=0: 
            if MaPlLib: p0.plot([                                          XX-SS*scaleP*pp[1],XX+SS*scaleP*pp[1]],[YY-SS*scaleP*pp[2],YY+SS*scaleP*pp[2]],'r-') # plot red from point to point
            else:              ffred.write("%12.4e  %12.4e\n%12.4e  %12.4e\n\n"%( XX-SS*scaleP*pp[1],YY-SS*scaleP*pp[2],  XX+SS*scaleP*pp[1],YY+SS*scaleP*pp[2] ))
        else:     
            if MaPlLib: p0.plot([                                          XX-SS*scaleP*pp[1],XX+SS*scaleP*pp[1]],[YY-SS*scaleP*pp[2],YY+SS*scaleP*pp[2]],'g-') # same as above but green
            else:              ffgre.write("%12.4e  %12.4e\n%12.4e  %12.4e\n\n"%( XX-SS*scaleP*pp[1],YY-SS*scaleP*pp[2],  XX+SS*scaleP*pp[1],YY+SS*scaleP*pp[2] ))
        if SS>Smax: ELaMax, Xmax, Ymax, Smax_, S2max = ElLabel, XX, YY, SS, pp[3]
#        if SS<Smin: ELaMin, Xmin, Ymin, Smin_, S2min = ElLabel, XX, YY, SS, pp[3]
        SS = pp[3]                                                          # smaller principal stress - signed
        if SS>=0: 
            if MaPlLib: p0.plot(                                          [XX-SS*scaleP*pp[2],XX+SS*scaleP*pp[2]],[YY+SS*scaleP*pp[1],YY-SS*scaleP*pp[1]],'r-') # plot red from point to point
            else:              ffred.write("%12.4e  %12.4e\n%12.4e  %12.4e\n\n"%( XX-SS*scaleP*pp[2],YY+SS*scaleP*pp[1],  XX+SS*scaleP*pp[2],YY-SS*scaleP*pp[1] ))
        else:     
            if MaPlLib: p0.plot(                                          [XX-SS*scaleP*pp[2],XX+SS*scaleP*pp[2]],[YY+SS*scaleP*pp[1],YY-SS*scaleP*pp[1]],'g-')# same as above but green
            else:              ffgre.write("%12.4e  %12.4e\n%12.4e  %12.4e\n\n"%( XX-SS*scaleP*pp[2],YY+SS*scaleP*pp[1],  XX+SS*scaleP*pp[2],YY-SS*scaleP*pp[1] ))
#        if SS>Smax: ELaMax, Xmax, Ymax, Smax_, S2max = ElLabel, XX, YY, SS, pp[0]
        if SS<Smin: ELaMin, Xmin, Ymin, Smin_, S2min = ElLabel, XX, YY, SS, pp[0]
        return ELaMax, ELaMin, Xmax, Ymax, Smax_, S2max, Xmin, Ymin, Smin_, S2min
    
    if not MaPlLib: import PyGnuplot as gp                                       # use gnuplot instead of matplotlib
    # store values of ndof per node -- if NodeResults exists use data from nodeout files, otherwise from actual displacement vector
    DisplList = NodalDisplList( NodeList, NodeResults)

#    # echo bond elements
#    if False:
#        FF = 0.
#        for i in ElemList:
#            if i.Type in ["B2D2E","B2D3E"]:
#                uElC, uElT, uElA = [], [], []
#                eC   = ElemList[i.iC]
#                if eC.Type=="CPS4_IGA": xyC = eC.ElCPLi                           # geometry control point coordinates 
#                else:                    xyC = eC.NodalCoordinates( NodeList, NoIndToCMInd )
#                XXC  = eC.FormX( i.rsC[0], i.rsC[1], 0.)                               
#                xyPC = dot(XXC,array(xyC))                                          # undeformed global coordinates of integration point
#                for j in eC.Inzi:                                                   # loop over nodes affected by element -- equivalent to lines before
#                    j_ = NoIndToCMInd[j]                                                   
#                    for k in range(len(DisplList[j_])):
#                        uElC += [DisplList[j_][k]]                                   # collects nodal displacements per element
#                val = None
#                if eC.Type=="CPS4_EFG":
#                    NNC = eC.ArbFormN( NodeList,NoIndToCMInd, CoorTree_,NoCoLitoNoLi, LevSetList, i.rsC[0], i.rsC[1], xyPC)
#                else:
#                    if eC.Type=="CPS4_XFEM" and i.DofE>8: val, valR, valT = eC.LevSetVal( i.rsC[0], i.rsC[1], LevSetList)
#                    NNC = eC.FormN( i.rsC[0], i.rsC[1], 0., val)
#                uC  = dot(NNC,uElC)
#                uC_ = dot(i.Rot,uC)
#                #
#                eT   = ElemList[i.iT]
#                xyT  = eT.NodalCoordinates( NodeList, NoIndToCMInd )                    # undeformed nodal coordinates
#                XXT  = eT.FormX( i.rsT[0], 0., 0.)                               
#                xyPT = dot(XXT,array(xyT))                                          # undeformed global coordinates of integration point
#                for j in eT.Inzi:                                                   # loop over nodes affected by element -- equivalent to lines before
#                    j_ = NoIndToCMInd[j]                                                   
#                    for k in range(len(DisplList[j_])):
#                        uElT += [DisplList[j_][k]]                                  # collects nodal displacements per element
#                NNT = eT.FormN( i.rsT[0], 0., 0., val)
#                uT  = dot(NNT,uElT)
#                uT_ = dot(i.Rot,uT)
#                for j in i.Inzi:
#                    j_ = NoIndToCMInd[j]                                                   
#                    for k in range(len(DisplList[j_])):
#                        uElA += [DisplList[j_][k]]                                  # collects nodal displacements per element
#                rL = zeros((i.DofE), dtype=float)
#                for j in range(i.nIntL):
#                    rst = SamplePoints[i.IntT,i.nInt,j]                             # local integration sampling point coordinates
#                    BB, JJ = i.FormB( rst[0], rst[1], rst[2], val )
#                    sl = dot(BB,uElA)
#                    tt, MatM, X = MatList[i.MatIp[j]].MatC( None, 1, 1., j, i, None, sl, None )                 # material for integration point
#                    tt = dot(MatM,sl)
#                    f = JJ*i.Geom*SampleWeight[i.IntT,i.nInt,j]
#                    rL = rL + f*dot(transpose(BB),tt)
#                    FF += MatM[0,0]*(uT_[0]-uC_[0])*f
#                    print('XXX', j, xyPC, xyPC-xyPT, '__', norm(uT_-uC_- sl), sl)
                    
    P1_, p1_, gi = [], [], 0                                                # gi counter for gnu plot windows
    # loop over all sections
    for s_, S in enumerate(SolSecs):
        if SolSecs[S].ElemTypes in [["S1D2"],["T1D2"],["B23"],["B23E"],["B21"],["B21E"] ]: continue            # this routine is not appropriate for this element types
        FlagPlotSec = False
        try:    
            ScaleSec =SCales[s_]
        except: 
            ScaleSec = [0]
            print("PostElem2D scaling factors for this section not provided, used 0",S)
        SigMax, SigMin, angle, maxX, minX = -9999, 9999, 0., -9999., 9999.
        ELaMax, ELaMin, Xmax, Ymax, Xmin, Ymin = 0, 0, 0, 0, 0, 0           # ELaMax element label
        Xmax_, Ymax_, Xmin_, Ymin_ = 0, 0, 0, 0
        if MaPlLib:
            if SolSecs[S].ElemTypes in [["SH4"],["SH3"]]:
                _, p1 = DefPlot("normal forces "  + SolSecs[S].Set+ " time "+"%6.3f"%TimeEl+ " scale "+"%7.4f"%ScaleSec[0]) 
                _, p2 = DefPlot("bending moments "+ SolSecs[S].Set+ " time "+"%6.3f"%TimeEl+ " scale "+"%7.4f"%ScaleSec[1]) 
                p2.axis('equal')
            elif SolSecs[S].ElemTypes in [["SB3"]]:
                _, p1 = DefPlot("bending moments "+ SolSecs[S].Set+ " time "+"%6.3f"%TimeEl+ " scale "+"%7.4f"%ScaleSec[0])
                scaleQ = 2.5*ScaleSec[1]
                _, p2 = DefPlot("shear forces "   + SolSecs[S].Set+ " time "+"%6.3f"%TimeEl+ " scale "+"%7.4f"%scaleQ) 
                p2.axis('equal')
            else:
                _, p1 = DefPlot("stresses "+ SolSecs[S].Set+ " time "+"%6.3f"%TimeEl+ " scale "+"%7.4f"%ScaleSec[0]) 
            p1.axis('equal')
#            P1_ += [P1]
            p1_ += [p1]
        else: 
            p1 = None
            ffg = open("tmp"+str(S)+".dat", 'w')                            # temporary file for gnuplot elements
            ffc = open("tmp"+str(S)+"c.dat", 'w')                           # temporary file for gnuplot cracks
            ffred = open("tmp"+str(S)+"red.dat", 'w')                       # temporary file for gnuplot principal stress tensile
            ffgre = open("tmp"+str(S)+"gre.dat", 'w')                       # temporary file for gnuplot principal stress compressive
            ffIP = open("tmp"+str(S)+"IP.dat", 'w')                         # temporary file for gnuplot integration points + color for stress level
            ffIP2= open("tmp"+str(S)+"IP2.dat", 'w')                        # temporary file for gnuplot integration points 2 + color for stress level
        xxMin, xxMax, yyMin, yyMax = 999., -999., 999., -999.
        # deformed mesh + cracks
        for jj in SolSecs[S].Elems:                                         # loop over element indices of Solid Section, element must currently not be active, as may have active for former output times
            i = ElemList[jj]
            if not i.Type in ["B2D2E","B2D3E"]:
                FlagPlotSec = True                                          # for later echo, see bottom
                key = str(i.Label)+i.Set+"0"                                # refers to 1st integration point to check whether this element exists at all
                xy = i.NodalCoordinates( NodeList, NoIndToCMInd )
                _, uxy =  ElementDisplList( i, DisplList, NoIndToCMInd, ScaleU)
                #
                linew=LiWiCrvThin
                if i.Type in ["T2D2","T2D2E","T2D3","T2D3E"]:
                    samp, xx, yy = [[-1.,0.,0.],[1.,0.,0.]], [], []         # samp: local sampling points for nodes, ZeroD for isogeom
                    linew=LiWiCrvMid
                elif i.Type in ["CPS3",'CPS3S',"CPE3",'CPE3S','CPS6','CPE6','CPS6S','CPE6S']:
                    samp, xx, yy = [[1., 0., 0.],[0., 1., 0.],[0., 0., 1.]], [], [] # samp: local sampling points for nodes CPS3
                elif i.Type in ["SB3","SH3"]:
                    samp, xx, yy = [[1., 0., 0.],[0., 1., 0.],[0., 0., 1.]], [], [] # samp: local sampling points for nodes SB3
                elif i.Type in ["SH4"]:
                    samp, xx, yy = [[-1., -1., 0.],[1., -1., 0.],[1., 1., 0.],[-1., 1., 0.]], [], []
                else:    
                    samp, xx, yy = [[-1.+ZeroD,-1.+ZeroD,0.],[1.-ZeroD,-1.+ZeroD,0.],[1.-ZeroD,1.-ZeroD,0.],[-1.+ZeroD,1.-ZeroD,0.]], [], []    # samp: local sampling points for nodes, ZeroD for isogeom
                for j in range(len(samp)):
                    r, s, t = samp[j][0], samp[j][1], samp[j][2] 
                    XX  = i.FormX_( r, s, t)
                    xyP = dot(XX,array(xy))                                 # undeformed coordinates
                    if i.Type in ["SH4","SH3"]: NN  = i.FormX_( r, s, 0.)
                    else:                       NN  = i.FormN( r, s, 0.)
                    xyP = xyP + dot(NN,array(uxy))                          # displaced coordinates
                    xx += [xyP[0]]
                    yy += [xyP[1]]
                    if xyP[0]<minX: minX=xyP[0]
                    if xyP[0]>maxX: maxX=xyP[0]
                xc, yc = mean(xx), mean(yy)                                 # coordinates of element center
                xx+= [xx[0]]                                                # to close the sequence of edges
                yy+= [yy[0]]                                                # to close the sequence of edges
                if min(xx) < xxMin: xxMin=min(xx)
                if max(xx) > xxMax: xxMax=max(xx)
                if min(yy) < yyMin: yyMin=min(yy)
                if max(yy) > yyMax: yyMax=max(yy)
                # start def
                def Plot2DCrackContour(CrXY0, CrXY1):
                    r0, s0, t0 = i.NLocCoorFromGlobCoor( xy, CrXY0)                     # local coordinate 'left' edge of crack
                    r1, s1, t1 = i.NLocCoorFromGlobCoor( xy, CrXY1)                     # local coordinate 'right' edge of crack
                    CrdXY0 = dot(i.FormN( r0, s0, t0),array(uxy))           # displacement 'left' edge of crack
                    CrdXY1 = dot(i.FormN( r1, s1, t1),array(uxy))           # displacement 'right' edge of crack
                    xx = [ CrXY0[0]+CrdXY0[0], CrXY1[0]+CrdXY1[0] ]
                    yy = [ CrXY0[1]+CrdXY0[1], CrXY1[1]+CrdXY1[1] ]
                    if MaPlLib: p1.plot(                                    [ xx[0],xx[1] ], [ yy[0],yy[1] ],color='gray')
                    else:              ffc.write("%8.4f  %8.4f\n%8.4f  %8.4f\n\n"%(  xx[0],yy[0],     xx[1],yy[1] ))
                # end of def
                if key in ElemResults:
                    if MaPlLib: 
                        p1.plot(                                  xx,yy,'--',color=Colors[s_],linewidth=linew) # matplotlib plot deformed element geometry
                        if i.Type in ["SH4","SH3","SB3"]: p2.plot(xx,yy,'--',color=Colors[s_],linewidth=linew)
#                        p1_[0].plot(xx,yy,'--',color=Colors[s_])           # 1st element set !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#                        try:    p1_[1].plot(xx,yy,'--',color=Colors[s_])           # 1st element set !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#                        except: pass
                    else:
                        for j in range(len(xx)): ffg.write("%8.4f  %8.4f\n"%(xx[j],yy[j]))  # gunplot write temp deformed mesh data 
                        ffg.write("\n")
                    # 2D SDA elements
                    if ElemResults[key][3] in ['CPS4S','CPE4S','CPS3S','CPE3S','CPS6S','CPE6S']: 
                        key = str(i.Label)+i.Set+"Cra"                      # key first crack
                        if key in ElemResults:
                            Data = ElemResults[key][5]
                            crX, crY, crNx, crNy = Data[0], Data[1], Data[2], Data[3]  # center x, center y, nx, ny
                            if MaPlLib: p1.plot([xc], [yc],'o',color='gray') # matplotlib mark cracked elements
                            xyC = i.FindGlobalDiscontinuityEdges( [crX,crY], [crNx,crNy], NodeList, NoIndToCMInd) # xyC: [[x, y of 1st edge],[x, y, of 2nd edge]]
                            Plot2DCrackContour( xyC[0], xyC[1] )
                        key = str(i.Label)+i.Set+"Cra2"                     # key second crack
                        if key in ElemResults:
                            Data = ElemResults[key][5]
                            crX, crY, crNx, crNy = Data[0], Data[1], Data[2], Data[3]  
                            if sqrt(crNx**2+crNy**2)>ZeroD:                 # crack is there
                                if MaPlLib: p1.plot([xc], [yc],'o',color='red')
                                xyC = i.FindGlobalDiscontinuityEdges( [crX,crY], [crNx,crNy], NodeList, NoIndToCMInd) # xyC: [[x, y of 1st edge],[x, y, of 2nd edge]]
                                Plot2DCrackContour( xyC[0], xyC[1] )
        # end of loop over deformed elements and cracks
#    label for elements, generally commented out
#                p1.text(xc,yc,"%i"%i.Label,ha='center',va='center',rotation=0.,color='black',fontsize=20) #7)
        # principal stresses on deformed geometry
        pDataX, pDataY = {}, {}
        Truss2DCoor, Truss2DStr, Truss2DStrBond,  = [], [], []
        for jj in SolSecs[S].Elems:                                         # loop over element indices of Solid Section
            i = ElemList[jj]
#                if i.Active and not i.Type in ["B2D2E","B2D3E","T1D2","S1D2"]:
            if not i.Type in ["B2D2E","B2D3E"]:
                for j in range(len(ScaleSec)): pDataX[j], pDataY[j] = [], []# storage for x,y of 2D plot per each scaling factor
                xy = i.NodalCoordinates( NodeList, NoIndToCMInd )
                _, uxy =  ElementDisplList( i, DisplList, NoIndToCMInd, ScaleU)
                # stresses in integration points
                if i.Type in ['T2D2','T2D2E','T2D3E','B23I','B23EI']: 
                    if i.cosA<0: angle =  90 - arcsin(i.sinA)*180/3.141593
                    else:        angle = -90 + arcsin(i.sinA)*180/3.141593
                    if angle < -91.: angle = angle + 180. 
#                    if i.Type in ['CPS4S','CPE4S','CPS3S','CPE3S']: ww = i.ww
                if   i.Type in ["SH3","SH4"]: Lis = i.Lists1()
                else:                         Lis = range(i.nIntL) 
                # loop over integration points of element
                for j in Lis:
                    r = SamplePoints[i.IntT,i.nInt-1,j][0]
                    s = SamplePoints[i.IntT,i.nInt-1,j][1]
                    t = SamplePoints[i.IntT,i.nInt-1,j][2]
                    key = str( i.Label)+i.Set+str(j)
                    if key in ElemResults: sig = DataToSig( ElemResults[key]) # key must not necessarily exist as actual ElemetList must not correspond to element data of previous output time
                    else:                  continue
                    XX  = i.FormX_( r, s, t)
                    xyP = dot(XX,array(xy))[0:2]                            # undeformed global coordinates of integration point
#                    
#                    print('YYY',j,xyP)
#                    
                    if i.Type in ["SH4","SH3"]: NN  = i.FormX_( r, s, 0.)
                    else:                       NN  = i.FormN( r, s, 0.)
                    xyP = xyP + dot(NN,array(uxy))[0:2]                     # displaced global coordinates of integration point
#                    CoLor = ColorTable[MatList[i.MatN].Type]
                    if MaPlLib:                                         
#                            p1.plot([xyP[0]], [xyP[1]],'x',color = 'darkred')       # matplotlib cross in integration point
#                            p1_[0].plot([xyP[0]], [xyP[1]],'x',color = 'darkred')  # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        pass
                    elif ElemResults[key][3] in ["T2D2","T2D2E","T2D3","T2D3E"]:    # for gnuplot
                        Truss2DCoor += [[xyP[0], xyP[1]]] 
                        Truss2DStr  += [sig[0]]                             # longitudinal stress
                        Truss2DStrBond  += [abs(sig[1])]                    # bond stress
                    # start def
                    def PlotRotLinesText( ind, sign_, pDataX, pDataY):      # seems to be used for truss elements only
                        if len(ScaleSec) != ind: raise NameError("ConFemPy3::PostElem2D: number of scaling factors does not match for this section",S,ScaleSec,ind)
                        phiC, phiS = i.cosA, i.sinA
                        sign_ = [1,-1,-1]
                        for jj in range(ind):
                            pDataX[jj] += [ xyP[0] - sign_[jj]*ScaleSec[jj]    *phiS*sig[jj] ]
                            pDataY[jj] += [ xyP[1] + sign_[jj]*ScaleSec[jj]    *phiC*sig[jj] ]
                            if sig[jj]<0.: colr = 'green'
                            else:          colr = 'red'
#                            p1.text(pDataX[jj][-1],pDataY[jj][-1],format(sig[jj],".2f"),ha='left',va='top',rotation=angle,color=Colors2[jj],fontsize='medium')
                            p1.text(pDataX[jj][-1],pDataY[jj][-1],format(sig[jj],".2f"),ha='left',va='top',rotation=angle,color=colr,fontsize='medium')
                    # end of def
                    if i.dim in [2,20,21]:                                  # 2D continuum elements, slab elements, shell elements
                        if i.Type == 'SB3': pp = PrinC( -sig[0], -sig[1], -sig[2] )  # to make positive values -- compression upper side -- 'green'
                        else:               pp = PrinC(  sig[0],  sig[1],  sig[2] )
                        ELaMax_, ELaMin_, Xmax_, Ymax_, Smax, S2max_, Xmin_, Ymin_, Smin, S2min_ = PloPrin( pp, ScaleSec[0], xyP[0], xyP[1], p1, SigMax, SigMin, i.Label, ELaMax, ELaMin, Xmax_, Ymax_, Xmin_, Ymin_)
                        SigMax, ELaMax, Xmax, Ymax, S2max = Smax, ELaMax_, Xmax_, Ymax_, S2max_ 
                        SigMin, ELaMin, Xmin, Ymin, S2min = Smin, ELaMin_, Xmin_, Ymin_, S2min_
                        if i.Type in ['SH3','SH4']:                         # for bending moments
                            pp = PrinC( sig[3], sig[4], sig[5] )
                            ELaMax_, ELaMin_, Xmax_, Ymax_, Smax, S2max_, Xmin_, Ymin_, Smin, S2min_ = PloPrin( pp, ScaleSec[1], xyP[0], xyP[1], p2, SigMax, SigMin, i.Label, ELaMax, ELaMin, Xmax_, Ymax_, Xmin_, Ymin_)
                        if i.Type == 'SB3':                                 # for shear forces
                            xS = (NodeList[NoIndToCMInd[i.Inzi[0]]].XCo+NodeList[NoIndToCMInd[i.Inzi[1]]].XCo+NodeList[NoIndToCMInd[i.Inzi[2]]].XCo)/3.
                            yS = (NodeList[NoIndToCMInd[i.Inzi[0]]].YCo+NodeList[NoIndToCMInd[i.Inzi[1]]].YCo+NodeList[NoIndToCMInd[i.Inzi[2]]].YCo)/3.
                            qx, qy = sig[3], sig[4]
                            if qx>0: p2.plot([xS,xS+scaleQ*qx],[yS,yS],'-',color='red')
                            else:    p2.plot([xS,xS+scaleQ*qx],[yS,yS],'-',color='green')
                            if qy>0: p2.plot([xS,xS],[yS,yS+scaleQ*qy],'-',color='red')
                            else:    p2.plot([xS,xS],[yS,yS+scaleQ*qy],'-',color='green')
                            p2.plot([xS],[yS],'o',color='gray')
                    elif i.Type in ['T2D2','T2D3']:
                        if MaPlLib: PlotRotLinesText( 1, [1], pDataX, pDataY)
                    elif i.Type in ['T2D2E','T2D3E']:
                        if MaPlLib: PlotRotLinesText( 3, [1, -1, -1], pDataX, pDataY)
                    elif i.Type in ['B23I','B23','B23EI','B23E']:           # result types see assignment for sig: low and up stress, long. and lat. bond stress
                        if MaPlLib:
                            if   i.Type in ['B23I','B23EI']: nData = 4
                            elif i.Type in ['B23','B23E']:  nData = 2
                            if len(ScaleSec) != nData: raise NameError("ConFemPy3::PostElem2D: number of scaling factors does not match for this section (4 required)",SolSecs[S].Set,ScaleSec)
                            phiC, phiS = i.cosA, i.sinA
                            sign_ = [1,1,-1,1]
                            for jj in range(nData):
                                pDataX[jj] += [ xyP[0] - sign_[jj]*ScaleSec[jj]    *phiS*abs(sig[jj]) ]
                                pDataY[jj] += [ xyP[1] + sign_[jj]*ScaleSec[jj]    *phiC*abs(sig[jj]) ]
                                p1.text(pDataX[jj][-1],pDataY[jj][-1],format(sig[jj],".2f"),ha='left',va='top',rotation=angle,color=Colors2[jj],fontsize='medium')
                if MaPlLib:
                    for a in pDataX: p1.plot(pDataX[a],pDataY[a], color = Colors2[a])
        # end of element loop
                
        # gnuplot truss element data
        if len(Truss2DCoor)>0:                                          
            Truss2DStrDel = max(Truss2DStr) - min(Truss2DStr) 
            if (Truss2DStrDel)>ZeroD:
                for j in range(len(Truss2DCoor)): 
                    ffIP.write("%12.4e  %12.4e  %12.4e\n"%(Truss2DCoor[j][0],Truss2DCoor[j][1],Truss2DStr[j]))
            Truss2DStrDel = max(Truss2DStrBond) - min(Truss2DStrBond) 
            if (Truss2DStrDel)>ZeroD:
                for j in range(len(Truss2DCoor)): 
                    ffIP2.write("%12.4e  %12.4e  %12.4e\n"%(Truss2DCoor[j][0],Truss2DCoor[j][1],Truss2DStrBond[j]))
        # gnuplot
        if not MaPlLib:
            ffg.close()
            ffc.close()
            ffred.close()
            ffgre.close()
            ffIP.close()
            ffIP2.close()
            gp.c('set term wxt '+str(gi))
            gp.c('set size square')
            gp.c('set grid')
            gp.c('unset key')
            gp.c('set title "Set '+i.Set+' -- Time '+str(TimeEl)+' -- Scale '+str(ScaleU)+' " ')
            dxx, dyy = max(1.,xxMax-xxMin), max(1.,yyMax-yyMin)
            gp.c('set xrange['+str(xxMin-0.08*dxx)+':'+str(xxMax+0.08*dxx)+']')
            gp.c('set yrange['+str(yyMin-0.08*dyy)+':'+str(yyMax+0.08*dyy)+']')
            gp.c('plot "tmp'+str(S)+'.dat" with lines linestyle 1 linecolor rgb "#0060ad" linetype 1 linewidth 0.5 ')              # deformed elements
            if os.path.getsize('tmp'+str(S)+'c.dat')>0:   gp.c('replot "tmp'+str(S)+'c.dat" with lines linestyle 1 linecolor rgb "gray30" ') # 2D crack contours 
            if os.path.getsize('tmp'+str(S)+'red.dat')>0: gp.c('replot "tmp'+str(S)+'red.dat" with lines linestyle 1 linecolor rgb "red" ') # tensile principal stresses
            if os.path.getsize('tmp'+str(S)+'gre.dat')>0: gp.c('replot "tmp'+str(S)+'gre.dat" with lines linestyle 1 linecolor rgb "dark-green" ') # commpressive principal stresses
            if os.path.getsize('tmp'+str(S)+'IP.dat')>0:  gp.c('replot "tmp'+str(S)+'IP.dat" using 1:2:3 w p pointtype 7 pointsize 0.3 lc palette')
            gp.p(filename=pdir+'tmp'+str(S)+'.eps', width=20, height=12, fontsize=12, term='wxt')
            if os.path.getsize('tmp'+str(S)+'IP2.dat')>0:
                gi += 1
                gp.c('set term wxt '+str(gi))
                gp.c('set size square')
                gp.c('set grid')
                gp.c('unset key')
                gp.c('set title "Set '+i.Set+' -- Time ' +str(TimeEl)+' -- Scale '+str(ScaleU)+' " ')
                gp.c('plot "tmp'+str(S)+'.dat" with lines linestyle 1 linecolor rgb "#0060ad" linetype 1 linewidth 0.5 ')              # deformed elements
                gp.c('replot "tmp'+str(S)+'IP2.dat" using 1:2:3 w p pointtype 7 pointsize 0.3 lc palette')
                gp.p(filename=pdir+'tmp'+str(S)+'_.eps', width=20, height=12, fontsize=12, term='wxt')
#            gp.c('pause -1')
#            gp.c('a = system("read a; echo $a")')
            ti.sleep(2)
            _ = str(input('\ncontinue: '))
            os.remove('tmp'+str(S)+'.dat');os.remove('tmp'+str(S)+'c.dat');os.remove('tmp'+str(S)+'red.dat');os.remove('tmp'+str(S)+'gre.dat');os.remove('tmp'+str(S)+'IP.dat');os.remove('tmp'+str(S)+'IP2.dat'); 
        # matplotlib
        else:
#    label for nodes, generally commented out
#            for i in NodeList: 
#                #if len(i.dL)==2: 
#                p1.text(i.XCo+ScaleU*i.dL[0],i.YCo+ScaleU*i.dL[1],"%i"%(i.Label-0),ha='left',va='bottom',rotation=0.,color='red',fontsize=7)
            #
#            for p in p1_: p.plot([minX-0.05*abs(maxX-minX),maxX+0.05*abs(maxX-minX)],[0,0],color='lightgrey') # ???
            try:    
                if i.dim==2: Annot()                     # i from above
            except:
                pass
        gi += 1
        if FlagPlotSec: 
            print('PostElem2D process plot for section',S,SolSecs[S].ElemTypes,"time",TimeEl) # ?! key S and value Set might be identical 
    return 0
 
#def PostElem2D_( ElList, NodeList,NoIndToCMInd, MatList, f2, Scales, SE, SP, ElPlotTimes ):
#    def IniP(Text, Time, ElSet, proj):
#        if proj=='3d': 
#            P0 = plt.figure().add_subplot(111,title=Text+'Time '+Time+', Element Set '+ElSet, projection=proj)
#        else:
#            P0 = plt.figure().add_subplot(111,title=Text+'Time '+Time+', Element Set '+ElSet)
#            P0.axis('equal')
#            P0.grid()
#        P0.tick_params(axis='x', labelsize=FonSizAx) #yticks(fontsize=FonSizAx) 
#        P0.tick_params(axis='y', labelsize=FonSizAx) #xticks(fontsize=FonSizAx)
#        P0.set_xlabel('x')
#        P0.set_ylabel('y') 
#        return True, P0
#    def RotLocal(Local1x, Local1y, pp):                                 # coordinates of 1st local axis
#        phi = -arcsin(Local1y/sqrt(Local1x**2+Local1y**2))
#        if Local1x<0.: phi = pi-phi
#        pp1 =  pp[1]*cos(phi) + pp[2]*sin(phi)                          # plane transformation of vector into new coordinate system
#        pp2 = -pp[1]*sin(phi) + pp[2]*cos(phi)
#        return pp1, pp2
#    def PloPrin( scaleP, XX, YY, P0, ELaMax, ELaMin, Xmax, Ymax, Smax, S2max, Xmin, Ymin, Smin, S2min, ElLabel):
#        SS = pp[0]
#        if SS>=0: P0.plot([XX-SS*scaleP*pp[1],XX+SS*scaleP*pp[1]],[YY-SS*scaleP*pp[2],YY+SS*scaleP*pp[2]],'r-')     # positive values red
#        else:     P0.plot([XX-SS*scaleP*pp[1],XX+SS*scaleP*pp[1]],[YY-SS*scaleP*pp[2],YY+SS*scaleP*pp[2]],'g-')     # negative values green
#        if SS>Smax: ELaMax, Xmax, Ymax, Smax, S2max = ElLabel, XX, YY, SS, pp[3]
#        if SS<Smin: ELaMin, Xmin, Ymin, Smin, S2min = ElLabel, XX, YY, SS, pp[3]
#        SS = pp[3]
#        if SS>=0: P0.plot([XX-SS*scaleP*pp[2],XX+SS*scaleP*pp[2]],[YY+SS*scaleP*pp[1],YY-SS*scaleP*pp[1]],'r-')     # positive values red
#        else:     P0.plot([XX-SS*scaleP*pp[2],XX+SS*scaleP*pp[2]],[YY+SS*scaleP*pp[1],YY-SS*scaleP*pp[1]],'g-')     # negative values green
#        if SS>Smax: ELaMax, Xmax, Ymax, Smax, S2max = ElLabel, XX, YY, SS, pp[0]
#        if SS<Smin: ELaMin, Xmin, Ymin, Smin, S2min = ElLabel, XX, YY, SS, pp[0]
#        return ELaMax, ELaMin, Xmax, Ymax, Smax, S2max, Xmin, Ymin, Smin, S2min
#    def Annot( ):
#        if Xmax!=None:
#            P0.annotate(    'max s1 '+str(Smax)[0:7]+' s2 '+str(S2max)[0:7]+' '+'%6i'%(Elmax),xy=(Xmax, Ymax),xycoords='data',xytext=(0.40,0.87),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='medium')
#            if ElType=='SH4': 
#                P1.annotate('max s1 '+'%7.4f'%(Smax_)+' s2 '+'%7.4f'%(S2max_),xy=(Xmax_, Ymax_),xycoords='data',xytext=(0.50,0.87),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='large')
#        if Xmin!=None:
#            P0.annotate(    'min s1 '+str(Smin)[0:7]+' s2 '+str(S2min)[0:7]+' '+'%6i'%(Elmin),xy=(Xmin, Ymin),xycoords='data',xytext=(0.40,0.84),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='medium')
#            if ElType=='SH4': 
#                P1.annotate('min s1 '+'%7.4f'%(Smin_)+' s2 '+'%7.4f'%(S2min_),xy=(Xmin_, Ymin_),xycoords='data',xytext=(0.50,0.84),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='large')
#        
#    ElPlotTimes = [float(i) for i in ElPlotTimes]
#    ESet, Tim, Xmax, Ymax, Xmin, Ymin, Elmax, Elmin = None, None, None, None, None, None, None, None
#    Xmax_, Ymax_, Xmin_, Ymin_ = None, None, None, None
#    ESet, Tim, TimesDone, FlagFoundEl = ' ', ' ', [], False
#    FlagStrain = False                                                      # Flag for strain in continuum based structural elements
#    z1 = f2.readline()
#    z2 = z1.split()
#    while z1!="":
#        if z2[0]=="Time":                                                   # TIME KEY FOUND
#            Time = z2[1]
#            if float(Time) in ElPlotTimes or len(ElPlotTimes)==0:           # plot all times in case len==0
#                if FlagFoundEl and len(TimesDone)>0: print('PostElem2D process plot for time ', TimesDone[-1])
#                TimesDone += [Time] 
#                FlagProc = True
#                FlagFoundEl = False
##                print("process plot for time ", Time)
#            else:
#                FlagProc = False
#        elif FlagProc and z2[0]=="El":                                      # ELEMENT KEY FOUND
#            ElType = z2[2]                                                  # key for element type
#            if ElType in ['CPE4','CPE4R','CPE3','CPS4','CPS4R','CPS3','CPE4S','CPE3S','CPS4S','CPS3S',
#                          'SB3','SH4','SH3',
#                          'T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
#                FlagFoundEl = True
#                ElSet = z2[3] 
#                from ConFemBasics import FindIndexByLabel                                              # key for element set
#                Elem  = ElList[FindIndexByLabel(ElList, int(z2[1]))]
#                from ConFemMat import ElasticLT
#                if isinstance( MatList[Elem.MatN], ElasticLT) or Elem.ShellRCFlag: FlagStrain=True
#                NewPlot = ElSet!=ESet or Time!=Tim
#                xN = [ NodeList[NoIndToCMInd[ii]].XCo for ii in Elem.Inzi]  # build list of element nodal x coordinates
#                yN = [ NodeList[NoIndToCMInd[ii]].YCo for ii in Elem.Inzi]  # build list of element nodal y coordinates
#                if ElType not in ['T2D2','T2D3','T2D2E','T2D3E']: 
#                    xN.append(NodeList[NoIndToCMInd[Elem.Inzi[0]]].XCo)     # close list of "
#                    yN.append(NodeList[NoIndToCMInd[Elem.Inzi[0]]].YCo)     # close list of "
#                if ElType in ['T2D3','T2D3E']:                              # additional bubble nodes have zero coordinates
#                    xN.pop(-1)
#                    yN.pop(-1)
#                if ElType in ['CPS4','CPS4R','CPS3','CPE4','CPE4R','CPE3','CPE4S','CPE3S','CPS4S','CPS3S',
#                              'SB3','SH4','SH3']:
#                    if NewPlot:                                             # new time or element set found
#                        Annot()
#                        if ElType == 'SB3': labels = ['moments: ','shear forces: ']
#                        else:               labels = ['normal forces: ','moments: ']
#                        IniFl, P0                                                     = IniP(labels[0],  Time, ElSet, None)   # normal forces
#                        if ElType=='SB3' or ElType=='SH4' or ElType=='SH3': IniFl, P1 = IniP(labels[1],  Time, ElSet, None)   # bending moments
#                        if ElType=='SH4' and FlagStrain:                    IniFl, P2 = IniP('strains: ',Time, ElSet, '3d')   # strains in 3d
#                        Smax, Smin, Smax_, Smin_, S2max, S2min, S2max_, S2min_ = -9999, 9999, -9999, 9999, None, None, None, None
#                        ScaleKey = ElSet+Time
#                        ESet, Tim, NewPlot = ElSet, Time, False
#                        scaleP = SE*Scales[ScaleKey][0]
#                        if ElType=='SH4' or ElType=='SH3': scaleP1 = SE*Scales[ScaleKey][1]
#                    P0.plot(xN,yN, 'b--')                               # undeformed ground view per element
#                    if ElType in ['SB3','SH4','SH3']: P1.plot(xN,yN, 'b--')# undeformed ground view per element
#                elif ElType in ['T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
#                    if NewPlot:                                         # new time or element set found
#                        IniFl, P4 = IniP( 'truss ', Time, ElSet, None)  # bar forces
#                        ESet, Tim, NewPlot = ElSet, Time, False
#                    P4.plot(xN,yN, 'ro')                                    # undeformed ground view per nodes of element
#                    P4.plot(xN,yN, 'b--')                               # undeformed ground view per element
#        elif FlagProc:                                                      # DATA FOUND
#            XX = float(z2[0])                                               # ip global x-coordinate
#            YY = float(z2[1])                                               # ip global y-coordinate
#            if ElType in ['CPS4','CPS4R','CPS3','CPE4','CPE4R','CPE3','CPE4S','CPE3S','CPS4S','CPS3S']:
#                pp = PrinC( float(z2[6]), float(z2[7]), float(z2[9])) 
#                Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min=PloPrin(scaleP,XX,YY,P0,Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min,Elem.Label)
#            elif ElType=='SB3':
#                pp = PrinC( float(z2[5]), float(z2[6]), float(z2[7]))
#                Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min=PloPrin(scaleP,XX,YY,P0,Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min,Elem.Label)
#            elif ElType=='SH4' or ElType=='SH3':
#                if z2[3]=='Z':                                          # integrated internal forces plotted to P0, P1
#                    xyLength = float(z2[12])**2+float(z2[13])**2        # length in xy-plane
#                    pp = PrinC( float(z2[4]), float(z2[5]), float(z2[6])) # n_xx, n_yy, n_xy
#                    if xyLength>ZeroD: pp[1], pp[2] = RotLocal( float(z2[12]), float(z2[13]), pp) # transformation of local axes into global
#                    Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min=PloPrin(scaleP,XX,YY,          P0 ,Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min,Elem.Label) # principal membrane forces
#                    pp = PrinC( float(z2[7]), float(z2[8]), float(z2[9])) # m_xx, m_yy, m_xy
#                    if xyLength>ZeroD: pp[1], pp[2] = RotLocal( float(z2[12]), float(z2[13]), pp) # transformation of local axes into global
#                    Elmax,Elmin,Xmax_,Ymax_,Smax_,S2max_,Xmin_,Ymin_,Smin_,S2min_=PloPrin(scaleP1,XX,YY, P1 ,Elmax,Elmin,Xmax_,Ymax_,Smax_,S2max_,Xmin_,Ymin_,Smin_,S2min_,Elem.Label) # principal moments
#                elif FlagStrain:                                        # plot data of every layer
#                    ZZ = float(z2[3])                                   # local local isoparametric t-coordinate
#                    if z2[4]=='R':                                      # reinforcement
#                        col='blue'
##                        P2.plot([XX],[YY],[ZZ],'o',color=col)           # single ip point plot 
#                    elif ZZ<-0.9 or ZZ>0.9:
##                    else:
#                        scaleX = SP*Scales[ScaleKey][2]
#                        pp = PrinC( float(z2[5]), float(z2[6]), 0.5*float(z2[7])) # principal strains concrete
#                        pp_ = PrinC( float(z2[8]), float(z2[9]), float(z2[10]))
##                        print('XXX',XX,YY,f"{ZZ:7.4f},  {pp[0]:7.4f},{pp[1]:7.4f},{pp[2]:7.4f},{pp[3]:7.4f},  {pp_[0]:8.4f},{pp_[3]:8.4f}")
#                        linesty = 'dotted'
#                        SS = pp[0]                                      # larger principal strain
#                        if SS > 0.1e-3: linesty = 'solid'
#                        if SS>=0: P2.plot([XX-SS*scaleX*pp[1],XX+SS*scaleX*pp[1]],[YY-SS*scaleX*pp[2],YY+SS*scaleX*pp[2]],[ZZ,ZZ],'r',linestyle=linesty)
#                        else:     P2.plot([XX-SS*scaleX*pp[1],XX+SS*scaleX*pp[1]],[YY-SS*scaleX*pp[2],YY+SS*scaleX*pp[2]],[ZZ,ZZ],'g',linestyle=linesty)
#                        linesty = 'dotted'
#                        SS = pp[3]
#                        if SS > 0.1e-3: linesty = 'solid'
#                        if SS>=0: P2.plot([XX-SS*scaleX*pp[2],XX+SS*scaleX*pp[2]],[YY+SS*scaleX*pp[1],YY-SS*scaleX*pp[1]],[ZZ,ZZ],'r',linestyle=linesty)
#                        else:     P2.plot([XX-SS*scaleX*pp[2],XX+SS*scaleX*pp[2]],[YY+SS*scaleX*pp[1],YY-SS*scaleX*pp[1]],[ZZ,ZZ],'g',linestyle=linesty)
#                        col='blue' # 'red'
#                        P2.plot([XX],[YY],[ZZ],'o',color=col,markersize=3)           # single ip point plot 
#            elif ElType in ['T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
#                if ElType in ['T2D2','T2D3','T2D2E','T2D3E']: FF = float(z2[3])
#                else:                                         FF = float(z2[4])
##                FF = float(z2[3]) # float(z2[-1]) # float(z2[4])                                           # z2 is current input line; 4 -> truss stress
##                Angle = arcsin(Elem.Geom[1,2]/sqrt(Elem.Geom[1,3]))*180/3.141593 #asin(Elem.Trans[0,1])*180/3.141593
#                Angle = arcsin(Elem.Geom[1,2]/Elem.LL)*180/3.141593 #asin(Elem.Trans[0,1])*180/3.141593
#                if FF>0: P4.text(XX,YY,format(FF,".2f"),ha='center',va='bottom',rotation=Angle,color='red',fontsize=FonSizAx)
#                else: P4.text(XX,YY,format(FF,".2f"),ha='center',va='bottom',rotation=Angle,color='green',fontsize=FonSizAx)
#            # slab shear
#            if ElType=='SB3':
#                xS = (NodeList[NoIndToCMInd[Elem.Inzi[0]]].XCo+NodeList[NoIndToCMInd[Elem.Inzi[1]]].XCo+NodeList[NoIndToCMInd[Elem.Inzi[2]]].XCo)/3.
#                yS = (NodeList[NoIndToCMInd[Elem.Inzi[0]]].YCo+NodeList[NoIndToCMInd[Elem.Inzi[1]]].YCo+NodeList[NoIndToCMInd[Elem.Inzi[2]]].YCo)/3.
#                qx = float(z2[8])
#                qy = float(z2[9])
#                scaleQ = 10.
#                if qx>0: P1.plot([xS,xS+scaleQ*qx],[yS,yS],'-',color='red')
#                else:    P1.plot([xS,xS+scaleQ*qx],[yS,yS],'-',color='green')
#                if qy>0: P1.plot([xS,xS],[yS,yS+scaleQ*qy],'-',color='red')
#                else:    P1.plot([xS,xS],[yS,yS+scaleQ*qy],'-',color='green')
#                P1.plot([xS],[yS],'o',color='red')
#        else:
#            pass
#        z1 = f2.readline()
#        z2 = z1.split()
#    if FlagFoundEl: 
#        print('PostElem2D process plot for time ', TimesDone[-1])
#    Annot()
#    return 0

# from former PostElem2D -- here specialized to plot layers of shells  
def PostElemSH4( ElList, NodeList,NoIndToCMInd, MatList, f2, Scales, SE, SP, ElPlotTimes, ScaleShellL ):
    def IniP(Text, Time, ElSet, proj):
        if proj=='3d': 
            P0 = plt.figure().add_subplot(111,title=Text+'Time '+Time+', Element Set '+ElSet, projection=proj)
        else:
            P0 = plt.figure().add_subplot(111,title=Text+'Time '+Time+', Element Set '+ElSet)
            P0.axis('equal')
            P0.grid()
        P0.tick_params(axis='x', labelsize=FonSizAx) #yticks(fontsize=FonSizAx) 
        P0.tick_params(axis='y', labelsize=FonSizAx) #xticks(fontsize=FonSizAx)
        P0.set_xlabel('x')
        P0.set_ylabel('y') 
        return True, P0
    def RotLocal(Local1x, Local1y, pp):                                 # coordinates of 1st local axis
        phi = -arcsin(Local1y/sqrt(Local1x**2+Local1y**2))
        if Local1x<0.: phi = pi-phi
        pp1 =  pp[1]*cos(phi) + pp[2]*sin(phi)                          # plane transformation of vector into new coordinate system
        pp2 = -pp[1]*sin(phi) + pp[2]*cos(phi)
        return pp1, pp2
    def PloPrin( scaleP, XX, YY, P0, ELaMax, ELaMin, Xmax, Ymax, Smax, S2max, Xmin, Ymin, Smin, S2min, ElLabel):
        SS = pp[0]
        if SS>=0: P0.plot([XX-SS*scaleP*pp[1],XX+SS*scaleP*pp[1]],[YY-SS*scaleP*pp[2],YY+SS*scaleP*pp[2]],'r-')     # positive values red
        else:     P0.plot([XX-SS*scaleP*pp[1],XX+SS*scaleP*pp[1]],[YY-SS*scaleP*pp[2],YY+SS*scaleP*pp[2]],'g-')     # negative values green
        if SS>Smax: ELaMax, Xmax, Ymax, Smax, S2max = ElLabel, XX, YY, SS, pp[3]
        if SS<Smin: ELaMin, Xmin, Ymin, Smin, S2min = ElLabel, XX, YY, SS, pp[3]
        SS = pp[3]
        if SS>=0: P0.plot([XX-SS*scaleP*pp[2],XX+SS*scaleP*pp[2]],[YY+SS*scaleP*pp[1],YY-SS*scaleP*pp[1]],'r-')     # positive values red
        else:     P0.plot([XX-SS*scaleP*pp[2],XX+SS*scaleP*pp[2]],[YY+SS*scaleP*pp[1],YY-SS*scaleP*pp[1]],'g-')     # negative values green
        if SS>Smax: ELaMax, Xmax, Ymax, Smax, S2max = ElLabel, XX, YY, SS, pp[0]
        if SS<Smin: ELaMin, Xmin, Ymin, Smin, S2min = ElLabel, XX, YY, SS, pp[0]
        return ELaMax, ELaMin, Xmax, Ymax, Smax, S2max, Xmin, Ymin, Smin, S2min
#    def Annot( ):
#        if Xmax!=None:
#            P0.annotate(    'max s1 '+str(Smax)[0:7]+' s2 '+str(S2max)[0:7]+' '+'%6i'%(Elmax),xy=(Xmax, Ymax),xycoords='data',xytext=(0.40,0.87),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='medium')
#            if ElType=='SH4': 
#                P1.annotate('max s1 '+'%7.4f'%(Smax_)+' s2 '+'%7.4f'%(S2max_),xy=(Xmax_, Ymax_),xycoords='data',xytext=(0.50,0.87),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='large')
#        if Xmin!=None:
#            P0.annotate(    'min s1 '+str(Smin)[0:7]+' s2 '+str(S2min)[0:7]+' '+'%6i'%(Elmin),xy=(Xmin, Ymin),xycoords='data',xytext=(0.40,0.84),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='medium')
#            if ElType=='SH4': 
#                P1.annotate('min s1 '+'%7.4f'%(Smin_)+' s2 '+'%7.4f'%(S2min_),xy=(Xmin_, Ymin_),xycoords='data',xytext=(0.50,0.84),textcoords='figure fraction',arrowprops=dict(arrowstyle="->"),fontsize='large')
#        
    ElPlotTimes = [float(i) for i in ElPlotTimes]
    ESet, Tim, Xmax, Ymax, Xmin, Ymin, Elmax, Elmin = None, None, None, None, None, None, None, None
    Xmax_, Ymax_, Xmin_, Ymin_ = None, None, None, None
    ESet, Tim, TimesDone, FlagFoundEl = ' ', ' ', [], False
    FlagStrain = False                                                      # Flag for strain in continuum based structural elements
    z1 = f2.readline()
    z2 = z1.split()
    while z1!="":
        if z2[0]=="Time":                                                   # TIME KEY FOUND
            Time = z2[1]
            if float(Time) in ElPlotTimes or len(ElPlotTimes)==0:           # plot all times in case len==0
                if FlagFoundEl and len(TimesDone)>0: print('PostElem2D process plot for time ', TimesDone[-1])
                TimesDone += [Time] 
                FlagProc = True
                FlagFoundEl = False
#                print("process plot for time ", Time)
            else:
                FlagProc = False
        elif FlagProc and z2[0]=="El":                                      # ELEMENT KEY FOUND
            ElType = z2[2]                                                  # key for element type
            if ElType in ['CPE4','CPE4R','CPE3','CPS4','CPS4R','CPS3','CPE4S','CPE3S','CPS4S','CPS3S',
                          'SB3','SH4','SH3',
                          'T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
                FlagFoundEl = True
                ElSet = z2[3] 
                from ConFemBasics import FindIndexByLabel                                              # key for element set
                Elem  = ElList[FindIndexByLabel(ElList, int(z2[1]))]
                from ConFemMat import ElasticLT
                if isinstance( MatList[Elem.MatN], ElasticLT) or Elem.ShellRCFlag: FlagStrain=True
                NewPlot = ElSet!=ESet or Time!=Tim
                xN = [ NodeList[NoIndToCMInd[ii]].XCo for ii in Elem.Inzi]  # build list of element nodal x coordinates
                yN = [ NodeList[NoIndToCMInd[ii]].YCo for ii in Elem.Inzi]  # build list of element nodal y coordinates
                if ElType not in ['T2D2','T2D3','T2D2E','T2D3E']: 
                    xN.append(NodeList[NoIndToCMInd[Elem.Inzi[0]]].XCo)     # close list of "
                    yN.append(NodeList[NoIndToCMInd[Elem.Inzi[0]]].YCo)     # close list of "
                if ElType in ['T2D3','T2D3E']:                              # additional bubble nodes have zero coordinates
                    xN.pop(-1)
                    yN.pop(-1)
                if ElType in ['CPS4','CPS4R','CPS3','CPE4','CPE4R','CPE3','CPE4S','CPE3S','CPS4S','CPS3S',
                              'SB3','SH4','SH3']:
                    if NewPlot:                                             # new time or element set found
#                        Annot()
                        if ElType == 'SB3': labels = ['moments: ','shear forces: ']
                        else:               labels = ['normal forces: ','moments: ']
#                        IniFl, P0                                                     = IniP(labels[0],  Time, ElSet, None)   # normal forces
#                        if ElType=='SB3' or ElType=='SH4' or ElType=='SH3': IniFl, P1 = IniP(labels[1],  Time, ElSet, None)   # bending moments
                        if ElType=='SH4' and FlagStrain:                    IniFl, P2 = IniP('strains: ',Time, ElSet, '3d')   # strains in 3d
                        Smax, Smin, Smax_, Smin_, S2max, S2min, S2max_, S2min_ = -9999, 9999, -9999, 9999, None, None, None, None
                        ScaleKey = ElSet+Time
                        ESet, Tim, NewPlot = ElSet, Time, False
                        scaleP = 1. #SE*Scales[ScaleKey][0]
                        if ElType=='SH4' or ElType=='SH3': 
                            scaleP1 = 1. #SE*Scales[ScaleKey][1]
#                    P0.plot(xN,yN, 'b--')                               # undeformed ground view per element
#                    if ElType in ['SB3','SH4','SH3']: P1.plot(xN,yN, 'b--')# undeformed ground view per element
                elif ElType in ['T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
                    if NewPlot:                                         # new time or element set found
                        IniFl, P4 = IniP( 'truss ', Time, ElSet, None)  # bar forces
                        ESet, Tim, NewPlot = ElSet, Time, False
                    P4.plot(xN,yN, 'ro')                                    # undeformed ground view per nodes of element
                    P4.plot(xN,yN, 'b--')                               # undeformed ground view per element
        elif FlagProc:                                                      # DATA FOUND
            XX = float(z2[0])                                               # ip global x-coordinate
            YY = float(z2[1])                                               # ip global y-coordinate
            if ElType in ['CPS4','CPS4R','CPS3','CPE4','CPE4R','CPE3','CPE4S','CPE3S','CPS4S','CPS3S']:
                pp = PrinC( float(z2[6]), float(z2[7]), float(z2[9])) 
#                Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min=PloPrin(scaleP,XX,YY,P0,Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min,Elem.Label)
            elif ElType=='SB3':
                pp = PrinC( float(z2[5]), float(z2[6]), float(z2[7]))
#                Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min=PloPrin(scaleP,XX,YY,P0,Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min,Elem.Label)
            elif ElType=='SH4' or ElType=='SH3':
                if z2[3]=='Z':                                          # integrated internal forces plotted to P0, P1
                    pass
#                    xyLength = float(z2[12])**2+float(z2[13])**2        # length in xy-plane
#                    pp = PrinC( float(z2[4]), float(z2[5]), float(z2[6])) # n_xx, n_yy, n_xy
#                    if xyLength>ZeroD: pp[1], pp[2] = RotLocal( float(z2[12]), float(z2[13]), pp) # transformation of local axes into global
#                    Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min=PloPrin(scaleP,XX,YY,          P0 ,Elmax,Elmin,Xmax,Ymax,Smax,S2max,Xmin,Ymin,Smin,S2min,Elem.Label) # principal membrane forces
#                    pp = PrinC( float(z2[7]), float(z2[8]), float(z2[9])) # m_xx, m_yy, m_xy
#                    if xyLength>ZeroD: pp[1], pp[2] = RotLocal( float(z2[12]), float(z2[13]), pp) # transformation of local axes into global
#                    Elmax,Elmin,Xmax_,Ymax_,Smax_,S2max_,Xmin_,Ymin_,Smin_,S2min_=PloPrin(scaleP1,XX,YY, P1 ,Elmax,Elmin,Xmax_,Ymax_,Smax_,S2max_,Xmin_,Ymin_,Smin_,S2min_,Elem.Label) # principal moments
                elif FlagStrain:                                        # plot data of every layer
                    ZZ = float(z2[3])                                   # local local isoparametric t-coordinate
                    if z2[4]=='R':                                      # reinforcement
                        col='blue'
#                        P2.plot([XX],[YY],[ZZ],'o',color=col)           # single ip point plot 
                    elif ZZ<-0.9 or ZZ>0.9:
#                    else:
                        scaleX = ScaleShellL # SP*Scales[ScaleKey][2]
                        pp = PrinC( float(z2[5]), float(z2[6]), 0.5*float(z2[7])) # principal strains concrete
                        pp_ = PrinC( float(z2[8]), float(z2[9]), float(z2[10]))
#                        print('XXX',XX,YY,f"{ZZ:7.4f},  {pp[0]:7.4f},{pp[1]:7.4f},{pp[2]:7.4f},{pp[3]:7.4f},  {pp_[0]:8.4f},{pp_[3]:8.4f}")
                        linesty = 'dotted'
                        SS = pp[0]                                      # larger principal strain
                        if SS > 0.1e-3: linesty = 'solid'
                        if SS>=0: P2.plot([XX-SS*scaleX*pp[1],XX+SS*scaleX*pp[1]],[YY-SS*scaleX*pp[2],YY+SS*scaleX*pp[2]],[ZZ,ZZ],'r',linestyle=linesty)
                        else:     P2.plot([XX-SS*scaleX*pp[1],XX+SS*scaleX*pp[1]],[YY-SS*scaleX*pp[2],YY+SS*scaleX*pp[2]],[ZZ,ZZ],'g',linestyle=linesty)
                        linesty = 'dotted'
                        SS = pp[3]
                        if SS > 0.1e-3: linesty = 'solid'
                        if SS>=0: P2.plot([XX-SS*scaleX*pp[2],XX+SS*scaleX*pp[2]],[YY+SS*scaleX*pp[1],YY-SS*scaleX*pp[1]],[ZZ,ZZ],'r',linestyle=linesty)
                        else:     P2.plot([XX-SS*scaleX*pp[2],XX+SS*scaleX*pp[2]],[YY+SS*scaleX*pp[1],YY-SS*scaleX*pp[1]],[ZZ,ZZ],'g',linestyle=linesty)
                        col='blue' # 'red'
                        P2.plot([XX],[YY],[ZZ],'o',color=col,markersize=3)           # single ip point plot 
            elif ElType in ['T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
                if ElType in ['T2D2','T2D3','T2D2E','T2D3E']: FF = float(z2[3])
                else:                                         FF = float(z2[4])
#                FF = float(z2[3]) # float(z2[-1]) # float(z2[4])                                           # z2 is current input line; 4 -> truss stress
#                Angle = arcsin(Elem.Geom[1,2]/sqrt(Elem.Geom[1,3]))*180/3.141593 #asin(Elem.Trans[0,1])*180/3.141593
                Angle = arcsin(Elem.Geom[1,2]/Elem.LL)*180/3.141593 #asin(Elem.Trans[0,1])*180/3.141593
                if FF>0: P4.text(XX,YY,format(FF,".2f"),ha='center',va='bottom',rotation=Angle,color='red',fontsize=FonSizAx)
                else: P4.text(XX,YY,format(FF,".2f"),ha='center',va='bottom',rotation=Angle,color='green',fontsize=FonSizAx)
            # slab shear
            if ElType=='SB3':
                xS = (NodeList[NoIndToCMInd[Elem.Inzi[0]]].XCo+NodeList[NoIndToCMInd[Elem.Inzi[1]]].XCo+NodeList[NoIndToCMInd[Elem.Inzi[2]]].XCo)/3.
                yS = (NodeList[NoIndToCMInd[Elem.Inzi[0]]].YCo+NodeList[NoIndToCMInd[Elem.Inzi[1]]].YCo+NodeList[NoIndToCMInd[Elem.Inzi[2]]].YCo)/3.
                qx = float(z2[8])
                qy = float(z2[9])
                scaleQ = 10.
                if qx>0: P1.plot([xS,xS+scaleQ*qx],[yS,yS],'-',color='red')
                else:    P1.plot([xS,xS+scaleQ*qx],[yS,yS],'-',color='green')
                if qy>0: P1.plot([xS,xS],[yS,yS+scaleQ*qy],'-',color='red')
                else:    P1.plot([xS,xS],[yS,yS+scaleQ*qy],'-',color='green')
                P1.plot([xS],[yS],'o',color='red')
        else:
            pass
        z1 = f2.readline()
        z2 = z1.split()
    if FlagFoundEl: 
        print('PostElem2D process plot for time ', TimesDone[-1])
#    Annot()
    return 0

def PostNode( ElList, NodeList,NICM, VecU, SN, ft):
    LX, LY, LZ, LU = [], [], [], []
    for i in NodeList:
        LX.append(i.XCo)
        LY.append(i.YCo)
        LZ.append(i.ZCo)
    for i in VecU: 
        LU.append(fabs(i))
    Xmin, Xmax, Ymin, Ymax, Zmin, Zmax = min(LX), max(LX), min(LY), max(LY), min(LZ), max(LZ)
    D = max(Xmax-Xmin,Ymax-Ymin,Zmax-Zmin)
    U = max(LU)
    if U > ZeroD: 
        scale = SN*0.5*D/U
    else:
        print("PostNode   Zero Displacements")
        scale = 0.                                              # estimation for scaling factor
    P0 = plt.figure().add_subplot(111,title='Deformed mesh final step (scale '+'%7.4f'%(scale)+')')
    P0.axis('equal')
    P0.grid()
    P0.tick_params(axis='x', labelsize=FonSizAx) #yticks(fontsize=FonSizAx) 
    P0.tick_params(axis='y', labelsize=FonSizAx) #xticks(fontsize=FonSizAx)
    P0.set_xlabel('x')
    P0.set_ylabel('y') 
    ElSets, colI = {}, 0                                            # for different colors for different element sets
    for Elem in ElList:
        xS, yS, xP, yP, xN, yN = [], [], [], [], [], []
        Set = Elem.Set
        if Set not in ElSets:
            ElSets[Set] = colI
            colI = colI+1
        if Elem.Type=='SB3':
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo,NodeList[NICM[Elem.Inzi[2]]].XCo,NodeList[NICM[Elem.Inzi[0]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo,NodeList[NICM[Elem.Inzi[2]]].YCo,NodeList[NICM[Elem.Inzi[0]]].YCo]
            # plot labels
            lN = [NodeList[NICM[Elem.Inzi[0]]].Label,NodeList[NICM[Elem.Inzi[1]]].Label,NodeList[NICM[Elem.Inzi[2]]].Label]
            xS_ = (NodeList[NICM[Elem.Inzi[0]]].XCo+NodeList[NICM[Elem.Inzi[1]]].XCo+NodeList[NICM[Elem.Inzi[2]]].XCo)/3.
            yS_ = (NodeList[NICM[Elem.Inzi[0]]].YCo+NodeList[NICM[Elem.Inzi[1]]].YCo+NodeList[NICM[Elem.Inzi[2]]].YCo)/3.
            eL = "%i/%i"%(i,Elem.Label)
            plt.text(xS_,yS_,eL,ha='center',va='center')                # elements
            for xL, yL, lL in zip(xN,yN,lN): plt.text(xL, yL, lL,ha='right',va='top',color='red') # nodes
        elif Elem.Type=='CPE4' or Elem.Type=='CPS4':
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo,NodeList[NICM[Elem.Inzi[2]]].XCo,NodeList[NICM[Elem.Inzi[3]]].XCo,NodeList[NICM[Elem.Inzi[0]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo,NodeList[NICM[Elem.Inzi[2]]].YCo,NodeList[NICM[Elem.Inzi[3]]].YCo,NodeList[NICM[Elem.Inzi[0]]].YCo]
            xP = [NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[1]]].XCo+scale*VecU[Elem.DofI[1,0]],NodeList[NICM[Elem.Inzi[2]]].XCo+scale*VecU[Elem.DofI[2,0]],NodeList[NICM[Elem.Inzi[3]]].XCo+scale*VecU[Elem.DofI[3,0]],NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]]]
            yP = [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]],NodeList[NICM[Elem.Inzi[1]]].YCo+scale*VecU[Elem.DofI[1,1]],NodeList[NICM[Elem.Inzi[2]]].YCo+scale*VecU[Elem.DofI[2,1]],NodeList[NICM[Elem.Inzi[3]]].YCo+scale*VecU[Elem.DofI[3,1]],NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]]]
        elif Elem.Type=='CPE3' or Elem.Type=='CPS3':
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo,NodeList[NICM[Elem.Inzi[2]]].XCo,NodeList[NICM[Elem.Inzi[0]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo,NodeList[NICM[Elem.Inzi[2]]].YCo,NodeList[NICM[Elem.Inzi[0]]].YCo]
            xP = [NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[1]]].XCo+scale*VecU[Elem.DofI[1,0]],NodeList[NICM[Elem.Inzi[2]]].XCo+scale*VecU[Elem.DofI[2,0]],NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]]]
            yP = [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]],NodeList[NICM[Elem.Inzi[1]]].YCo+scale*VecU[Elem.DofI[1,1]],NodeList[NICM[Elem.Inzi[2]]].YCo+scale*VecU[Elem.DofI[2,1]],NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]]]
        elif Elem.Type=='SH4':
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo,NodeList[NICM[Elem.Inzi[2]]].XCo,NodeList[NICM[Elem.Inzi[3]]].XCo,NodeList[NICM[Elem.Inzi[0]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo,NodeList[NICM[Elem.Inzi[2]]].YCo,NodeList[NICM[Elem.Inzi[3]]].YCo,NodeList[NICM[Elem.Inzi[0]]].YCo]
            xP = [NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[1]]].XCo+scale*VecU[Elem.DofI[1,0]],NodeList[NICM[Elem.Inzi[2]]].XCo+scale*VecU[Elem.DofI[2,0]],NodeList[NICM[Elem.Inzi[3]]].XCo+scale*VecU[Elem.DofI[3,0]],NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]]]
            yP = [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]],NodeList[NICM[Elem.Inzi[1]]].YCo+scale*VecU[Elem.DofI[1,1]],NodeList[NICM[Elem.Inzi[2]]].YCo+scale*VecU[Elem.DofI[2,1]],NodeList[NICM[Elem.Inzi[3]]].YCo+scale*VecU[Elem.DofI[3,1]],NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]]]
        elif Elem.Type=='SH3':
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo,NodeList[NICM[Elem.Inzi[2]]].XCo,NodeList[NICM[Elem.Inzi[0]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo,NodeList[NICM[Elem.Inzi[2]]].YCo,NodeList[NICM[Elem.Inzi[0]]].YCo]
            xP = [NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[1]]].XCo+scale*VecU[Elem.DofI[1,0]],NodeList[NICM[Elem.Inzi[2]]].XCo+scale*VecU[Elem.DofI[2,0]],NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]]]
            yP = [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]],NodeList[NICM[Elem.Inzi[1]]].YCo+scale*VecU[Elem.DofI[1,1]],NodeList[NICM[Elem.Inzi[2]]].YCo+scale*VecU[Elem.DofI[2,1]],NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]]]
        elif Elem.Type in ['T1D2']:
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo]
            xP = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo]
            yP = [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[1]]].YCo+scale*VecU[Elem.DofI[1,0]]]
        elif Elem.Type in ['T2D2','T2D3','T2D2E','T2D3E','T3D2','T3D3','T3D2E','T3D3E']:
            xN = [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[1]]].XCo]
            yN = [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[1]]].YCo]
            xP = [NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[1]]].XCo+scale*VecU[Elem.DofI[1,0]]]
            yP = [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]],NodeList[NICM[Elem.Inzi[1]]].YCo+scale*VecU[Elem.DofI[1,1]]]
        elif Elem.Type=='S1D2':
            j=0
            k0 = Elem.DofI[j,0]
            k1 = Elem.DofI[j+1,0]
            xS += [NodeList[NICM[Elem.Inzi[j]]].XCo]
            yS += [NodeList[NICM[Elem.Inzi[j]]].YCo+scale*(VecU[k1]-VecU[k0])]
        elif Elem.Type in ['B23E','B23EI']:
            xN += [NodeList[NICM[Elem.Inzi[0]]].XCo,NodeList[NICM[Elem.Inzi[2]]].XCo]
            yN += [NodeList[NICM[Elem.Inzi[0]]].YCo,NodeList[NICM[Elem.Inzi[2]]].YCo]
            xP =  [NodeList[NICM[Elem.Inzi[0]]].XCo+scale*VecU[Elem.DofI[0,0]],NodeList[NICM[Elem.Inzi[2]]].XCo+scale*VecU[Elem.DofI[2,0]]]
            yP =  [NodeList[NICM[Elem.Inzi[0]]].YCo+scale*VecU[Elem.DofI[0,1]],NodeList[NICM[Elem.Inzi[2]]].YCo+scale*VecU[Elem.DofI[2,1]]]
        elif Elem.Type not in ['SB3','CPE4','CPS4','SH4', 'SH3', 'T1D2','S1D2','CPE3','CPS3','B23E',
                               'B2D2E','B2D3E','B3D2E','B3D3E']:
            for j in range(Elem.nNod):
                k0 = Elem.DofI[j,0]
                k1 = Elem.DofI[j,1]
                xN += [NodeList[NICM[Elem.Inzi[j]]].XCo]                    # undeformed 2D geometry
                yN += [NodeList[NICM[Elem.Inzi[j]]].YCo]                    # undeformed 2D geometry
                xP += [NodeList[NICM[Elem.Inzi[j]]].XCo+scale*VecU[Elem.DofI[j,0]]]
                yP += [NodeList[NICM[Elem.Inzi[j]]].YCo+scale*VecU[Elem.DofI[j,1]]]
        P0.plot(xN,yN, 'b--')                                               # undeformed geometry for current element
        if (len(xP)+len(yP))>0:                                             # deformed geometry for current element
            P0.plot(xP,yP, 'b-',linewidth=LiWiCrv)
            P0.plot(xP,yP, 'o', color=Colors[ElSets[Set]])
            if ft != None:
                print(Elem.Type,',',scale,',',xN,',',yN,',',xP,',',yP, file=ft )
        if len(xS)>0:                                                       # for S1D2 only
            P0.plot(xS,yS, 'b-',linewidth=LiWiCrv)
            P0.plot(xS,yS, 'o', color=Colors[ElSets[Set]])
    return 0

def PostElem3D( ElList, NodeList,NoIndToCMInd, ElemResults, Time):
    XList, YList, ZList, LisLen = [], [], [], 0
    nxLi,nyLi,nxyLi,mxLi,myLi,mxyLi,qxLi,qyLi, aaLi = [], [], [], [], [], [], [], [], [] # collect internal force / thickness data for all elements
    for i in NodeList:
        XList += [i.XCo]
        YList += [i.YCo]
        ZList += [i.ZCo]
    # 
    for elem in ElList: 
        if elem.Type in ['SH4','SH3']:
            offset, nRe = 0, 0                                              # currently not clear for what offset is, does this reinforcement work ???  uhc 220621
            if elem.ShellRCFlag: nRe = elem.Geom.shape[0]-2                 # number of reinforcement layers
            else:                nRe = 0
            nx, ny, nxy, qy, qx, mx, my, mxy, aa_ = 0., 0., 0., 0., 0., 0., 0., 0.,  0.
            Lis = elem.Lists1()                                             # indices for first integration points in base area      
            for j in Lis:
                nx_, ny_, nxy_, qx_, qy_, mx_, my_, mxy_, aa = elem.StressIntegration( j, offset, nRe, ElemResults, None) # uses internal forces from ElemResults
                nx += nx_  
                ny += ny_ 
                nxy+= nxy_  
                qx += qx_
                qy += qy_
                mx += mx_
                my += my_
                mxy+= mxy_  
                aa_ = aa_ + aa
            # end of base integration point loop - following for means of element 
            LisLen = len(Lis)                                               # number of integration points in base area
            nxLi.append( nx/LisLen)
            nyLi.append( ny/LisLen)
            nxyLi.append(nxy/LisLen)
            mxLi.append( mx/LisLen)
            myLi.append( my/LisLen)
            mxyLi.append(mxy/LisLen)
            qxLi.append( qx/LisLen)
            qyLi.append( qy/LisLen)
            aaLi.append( aa_/LisLen)
    # end of element loop
    # plot whole stuff
    if LisLen>0:
        TolP = 1.0e-9
        maxX, minX, maxY, minY, maxZ, minZ = max(XList), min(XList), max(YList), min(YList), max(ZList), min(ZList)
        plotVal3D(aaLi, 'thickness', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        if max(map(abs,nxLi)) > TolP: plotVal3D(nxLi, 'n_x', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        if max(map(abs,nyLi)) > TolP: plotVal3D(nyLi, 'n_y', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        if max(map(abs,nxyLi))> TolP: plotVal3D(nxyLi,'n_xy',minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        plotVal3D(mxLi, 'm_x', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        plotVal3D(myLi, 'm_y', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        plotVal3D(mxyLi,'m_xy',minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        plotVal3D(qxLi, 'q_x', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
        plotVal3D(qyLi, 'q_y', minX,maxX,minY,maxY,minZ,maxZ,ElList,NodeList,NoIndToCMInd, Time)
    #
    return LisLen

def plotVal3D( valList,Label, minX,maxX,minY,maxY,minZ,maxZ, ElList, NodeList,NoIndToCMInd, Time):
    delX, delY, delZ = maxX-minX, maxY-minY, maxZ-minZ
    delRef = max(delX, delY, delZ)
    delR2  = 0.5*delRef
    meanX, meanY, meanZ = 0.5*(minX+maxX), 0.5*(minY+maxY), 0.5*(minZ+maxZ) 
    ColM = plt.get_cmap('brg')
    fig = pl.figure() #figsize=(15.,10.))
    fig.text( 0.05, 0.93, Label+' time '+str(Time), size='x-large')
    maxVal = max(valList)
    minVal = min(valList)
    if abs(maxVal-minVal)<1.e-6: 
        aVal = abs(maxVal)
        maxVal = maxVal + 0.2*aVal
        minVal = minVal - 0.2*aVal
    norm_ = colors.Normalize(vmin=minVal, vmax=maxVal)
    ax1, orie = fig.add_axes([0.90, 0.05, 0.02, 0.8]), 'vertical'
#    ax1, orie = fig.add_axes([0.05, 0.05, 0.9, 0.02]), 'horizontal' # left, bottom, width, height] where all quantities are in fractions of figure width and height
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap=ColM,  norm=norm_, orientation=orie)
    ax = a3.Axes3D(fig)
    for i in range(len(ElList)):
        elem, CoordList = ElList[i], []
        for j in range(len(elem.Inzi)):
            node = NodeList[elem.Inzi[j]]
            ni = NoIndToCMInd[elem.Inzi[j]]
            node = NodeList[ni]
            CoordList += [[node.XCo, node.YCo, node.ZCo]]
        tri = a3.art3d.Poly3DCollection([CoordList])
        tri.set_color(ColM((valList[i] - minVal)/(maxVal-minVal)))                      # --> valList
        tri.set_edgecolor('k')
        ax.add_collection3d(tri)
    ax.set_xlim3d( meanX-delR2, meanX+delR2)
    ax.set_ylim3d( meanY-delR2, meanY+delR2)
#    if abs(minZ)<0.01: minZ = -0.1
#    if     maxZ <0.01: maxZ =  0.1 
    ax.set_zlim3d(meanZ-delR2, meanZ+delR2)
    ax.set_xlabel('x',fontsize='x-large')
    ax.set_ylabel('y',fontsize='x-large')
    ax.set_zlabel('z',fontsize='x-large')
    ax.tick_params(axis='x', labelsize='x-large')
    ax.tick_params(axis='y', labelsize='x-large')# ,labelcolor='white')
    ax.tick_params(axis='z', labelsize='x-large')#small',labelcolor='white')
    ax.set_xticks([minX,meanX,maxX])
    ax.set_xticklabels( ["%.2f"%(minX),"%.2f"%(meanX),"%.2f"%(maxX)] )
    ax.set_yticks([minY,meanY,maxY])
    ax.set_yticklabels( ["%.2f"%(minY),"%.2f"%(meanY),"%.2f"%(maxY)] )
    ax.set_zticks([minZ,meanZ,maxZ])
    ax.set_zticklabels( ["%.2f"%(minZ),"%.2f"%(meanZ),"%.2f"%(maxZ)] )
    ax.set_aspect('auto') #,'box', 'equal')
    ax.grid()
    return 0

def PostNode3D( ElList,NodeList,NoIndToCMInd, NodeResults,VecU, Time, ScaleDis):
    XList, YList, ZList, DisElemList = [], [], [], []
    for i in NodeList:
        XList += [i.XCo]
        YList += [i.YCo]
        ZList += [i.ZCo]
    xMin, xMax, XM = np.min(XList), np.max(XList), np.mean(XList)
    yMin, yMax, YM = np.min(YList), np.max(YList), np.mean(YList)
    zMin, zMax, ZM = np.min(ZList), np.max(ZList), np.mean(ZList)
    delX, delY, delZ = xMax-xMin, yMax-yMin, zMax-zMin
    delRef = max(delX, delY, delZ)
    delR2  = 0.5*delRef
    xMean, yMean, zMean = 0.5*(xMin+xMax), 0.5*(yMin+yMax), 0.5*(zMin+zMax) 
    # for measure of absolute displacement
    for elem in ElList:                                                     # loop over elements
        DisList = []
        for o in elem.Inzi:
            ni = NoIndToCMInd[o]
            node = NodeList[ni]
            if len(VecU)==0:
                d = NodeResults[node.Label][2]
                if   elem.Type in ['SH4','SH3']: DisList += [ sqrt(d[0]**2 + d[1]**2 + d[2]**2) ]
                elif elem.Type=='SB3':           DisList += [ sqrt(d[0]**2) ]
            else:
                DofInd = node.GlobDofStart
                if   elem.Type in ['SH4','SH3']: DisList += [sqrt(VecU[DofInd]**2+VecU[DofInd+1]**2+VecU[DofInd+2]**2)]
                elif elem.Type=='SB3':           DisList += [VecU[DofInd]]       
                pass
        DisElemList += [np.mean(DisList)]                                   # this is always positive
    dMax = max(DisElemList)
    #
    scale_ = 10.                                                            # 20.  # displayed deformation should be roughly 1/scale_ of parts dimensions
    scale = ScaleDis*max(XM,YM,ZM)/(dMax*scale_)
    ColM = plt.get_cmap('summer')
    fig = pl.figure()
#    fig.text( 0.05, 0.93, 'deformed mesh - scale: {:7.3f} '.format(scale)+text, size='x-large') # , ha='right')
    fig.text( 0.05, 0.95, f"deformed mesh time {Time:7.3f} - scale {scale:7.3f}", size='x-large') # , ha='right')
    norm_ = colors.Normalize(vmin=0., vmax=dMax)
    ax1, orie = fig.add_axes([0.90, 0.05, 0.02, 0.8]), 'vertical'           # left, bottom, width, height] where all quantities are in fractions of figure width and height
#    ax1, orie = fig.add_axes([0.05, 0.05, 0.9, 0.02]), 'horizontal' # left, bottom, width, height] where all quantities are in fractions of figure width and height
    cb1 = mpl.colorbar.ColorbarBase(ax1, cmap=ColM, norm=norm_, orientation=orie)
    ax = a3.Axes3D(fig)
    #
    for i, elem in enumerate( ElList ):
        CoordList  = []
        for j in elem.Inzi:
            node = NodeList[ NoIndToCMInd[j] ]
            if len(VecU) == 0:
                nL = node.Label
                if nL in NodeResults: noRe = NodeResults[nL][2]
                else:                 raise NameError("ConFemPost::PostNode3D: node results not available",nL)
                if   elem.Type in ['SH4','SH3']: CoordList += [[node.XCo+scale*noRe[0],  node.YCo+scale*noRe[1],  node.ZCo+scale*noRe[2]]]
                elif elem.Type=='SB3':           CoordList += [[node.XCo-xMin, node.YCo-yMin, node.ZCo-zMin+scale*noRe[0]]]
                else: raise NameError ("ConFemPost::PostNode3D: wrong element type",elem.Label,elem.Type)
            else:
                DofInd = node.GlobDofStart
                if   elem.Type in ['SH4','SH3']: CoordList += [[node.XCo+scale*VecU[DofInd], node.YCo+scale*VecU[DofInd+1], node.ZCo+scale*VecU[DofInd+2]]]
                elif elem.Type=='SB3':           CoordList += [[node.XCo-xMin, node.YCo-yMin, node.ZCo-zMin+scale*VecU[DofInd]]]
        tri = a3.art3d.Poly3DCollection([CoordList])
        tri.set_color(ColM( DisElemList[i]/dMax ))
        tri.set_edgecolor('k')
        ax.add_collection3d(tri)
    #
    ax.set_xlim3d( xMean-delR2, xMean+delR2 )
    ax.set_ylim3d( yMean-delR2, yMean+delR2 )
    ax.set_zlim3d( zMean-delR2, zMean+delR2 )
    ax.set_xlabel('x',fontsize='x-large')
    ax.set_ylabel('y',fontsize='x-large')
    ax.set_zlabel('z',fontsize='x-large')
    ax.tick_params(axis='x', labelsize='x-large')
    ax.tick_params(axis='y', labelsize='x-large')
    ax.tick_params(axis='z', labelsize='x-large')
    ax.set_xticks([xMin,xMean,xMax])
    ax.set_xticklabels( ["%.2f"%(xMin),"%.2f"%(xMean),"%.2f"%(xMax)] )
    ax.set_yticks([yMin,yMean,yMax])
    ax.set_yticklabels( ["%.2f"%(yMin),"%.2f"%(yMean),"%.2f"%(yMax)] )
    ax.set_zticks([zMin,zMean,zMax])
    ax.set_zticklabels( ["%.2f"%(zMin),"%.2f"%(zMean),"%.2f"%(zMax)] )
    ax.set_aspect('auto')
    ax.grid()
    return 0

def PlotHist( f5, WrNodes ):
    if WrNodes[0][3]=='p-u':   label='load-displacement node '+str(WrNodes[0][0])
    elif WrNodes[0][3]=='p-t': label='load-time node '+str(WrNodes[0][0])
    elif WrNodes[0][3]=='u-t': label='displacement-time (node '+str(WrNodes[0][0])+')'
    elif WrNodes[0][3]=='b-t': label='reaction force-time (node '+str(WrNodes[0][0])+')'
    elif WrNodes[0][3]=='b-u': label='reaction force-displacement (node '+str(WrNodes[0][0])+')'
    else: raise NameError("ConFemInOut:PlotNodes: unknown key")
    sc1, sc2 = 1., 1.
#    sc1, sc2 = -1., -1.
    L1, L2 = [0.], [0.]
    z1 = f5.readline()                                                      # 1st input line
    z2 = z1.split()
    while z1!="":
#        zLen = len(z2)                                                    
        zLen = len(z2)-2                                                    # for adding step, counter to end of line
        if (zLen % 2) != 0: raise NameError("ConFemInOut::PlotNodes: input line inconsistent",zLen,z2)
        sum_ = 0.
#        for i in range(int(zLen/2)): sum_ += float(z2[2*i-1])
        for i in range(int(zLen/2)): sum_ += float(z2[2*i+1])
        L1 += [sc1*float(z2[0])]
        L2 += [sc2*sum_]
        z1 = f5.readline()                                                  # read next line
        z2 = z1.split()
    PP = plt.figure()
    P0 = PP.add_subplot(111,title=label) 
    P0.set_title(label,fontsize=FonSizTi)
    P0.plot(L1,L2,linewidth=LiWiCrv)
    P0.set_ylim(1.05*min(L2),1.05*max(L2))
    P0.tick_params(axis='x', labelsize=FonSizAx)
    P0.tick_params(axis='y', labelsize=FonSizAx) #'x-large')
    P0.grid()
    PP.autofmt_xdate()
    return 0

# store values of ndof per node 
def NodalDisplList( NodeList, NodeResults):
    DisplList = [] 
    for i in NodeList:
        key = i.Label
        nDof = NodeResults[key][1]
        Data = NodeResults[key][2]
        dL = []
#        for j in range(nDof): dL += [Data[j]]
        for j in range(min(nDof,3)): dL += [Data[j]]                        # use coordinates only
        DisplList += [dL]
        i.dL = dL                                                           # for label text on deformed structure
    return DisplList
# store values of ndof per element 
def ElementDisplList( i, DisplList, NoIndToCMInd, ScaleU):
    uEl, uxy = [], []                             
    for j in i.Inzi:                                                        # loop over nodes affected by element
        j_ = NoIndToCMInd[j]
        for k in range(len(DisplList[j_])):
            uEl += [DisplList[j_][k]]                                       # collects nodal displacements per element
            uxy += [ScaleU*DisplList[j_][k]]                                # collects scaled nodal displacements per element
    return uEl, uxy

def PrinC( xx, yy, xy ):                                                    # calculation of principal values
    if ZeroD < abs(xy):
        h1 = xx+yy
        h2 = sqrt((xx-yy)**2+4*xy**2);
        s1 = .5*h1+.5*h2
        s2 = .5*h1-.5*h2
        h = (s1-xx)/xy
        L = sqrt(h**2+1)
        n11 = 1./L
        n12 = h/L
        h = (s2-xx)/xy
        L = sqrt(h**2+1)
        n21 = 1./L
        n22 = h/L 
    else:
        s1 = xx
        n11 = 1.
        n12 = 0
        s2 = yy
        n21 = 0
        n22 = 1. 
    return( [s1, n11, n12, s2, n21, n22] )

def ReadLine(ff):
    z1 = ff.readline()
    z2 = z1.strip()
    return z2.split(',')

def ReadResultsFromFile( ff, ffN):
    ElemResults, NodeResults, EndFlag, EndFlagN  = {}, {}, False, False
    # read from elemout_ -- current line should be time line
    z1 = ff.readline()
    z2 = z1.strip()
    z3 = z2.split(',')
    if z3[0]!="Time": raise NameError("ConFemPostProcStandAlone: something wrong with elemout_")
    else:             Time = float(z3[1])
    while z1!="":
        pos= ff.tell()
        z1 = ff.readline()
        z2 = z1.strip()
        z3 = z2.split(',')
        if z1   =="":                   break
        if z3[0]=='Time': ff.seek(pos); break
        key = z3[0].strip()+z3[1].strip()+z3[4].strip()                     # key is label + elset + integration point
        Data = []
        ll = len(z3) 
        for i in z3[5:ll+1]:                                                # starts with 6th position (-> index 5) of *.elemout_.txt
            try:    Data += [float(i)]                                      # maybe colon as last item
            except: pass                               
        ElemResults[key] = [z3[0].strip(), z3[1].strip(), z3[2].strip(), z3[3].strip(), z3[4].strip(), Data]
    if z1=="": EndFlag = True
    # read from nodeout_
    z1 = ffN.readline()
    z2 = z1.strip()
    z3N= z2.split(',')
    if z3N[0]!="Time": raise NameError("ConFemPostProcStandAlone: something wrong with nodeout_ ",z3N)
    else:             TimeN = float(z3N[1])
    while z1!="":
        pos= ffN.tell()
        z1 = ffN.readline()
        z2 = z1.strip()
        z3N= z2.split(',')
        if z1    =="":                    break
        if z3N[0]=='Time': ffN.seek(pos); break
        key  = int(z3N[0])                                                  # int node label, should be unique
        nDof = int(z3N[4])                                                  # number of degrees of freedom
        Data = []
        for i in z3N[5:5+nDof]: Data += [float(i)]
        NodeResults[key] = [ key, nDof, Data]
    if z1=="": EndFlagN = True
    #
    return EndFlag, EndFlagN, Time, TimeN, ElemResults, NodeResults

def PickleLoad( FilDir, FilName, StepCounter):
    import pickle
#    from scipy.spatial import KDTree
    import scipy.spatial as spatial
#    from os import path as pth
    if pth.isfile(FilDir+FilName+".pkl"):
        fd = open(FilDir+FilName+'.pkl', 'rb')                              # has to be in sync with pickle.dumo in ConFem, ConSimFem
        NodeList=pickle.load(fd);ElList=pickle.load(fd);MatList=pickle.load(fd);StepList=pickle.load(fd);N=pickle.load(fd);WrNodes=pickle.load(fd);LineS=pickle.load(fd);FlElasticLT=pickle.load(fd);\
            VecU=pickle.load(fd);VecC=pickle.load(fd);VecI=pickle.load(fd);VecP=pickle.load(fd);VecP0=pickle.load(fd);VecP0old=pickle.load(fd);VecBold=pickle.load(fd);VecT=pickle.load(fd);VecS=pickle.load(fd);\
            VeaU=pickle.load(fd);VevU=pickle.load(fd);VeaC=pickle.load(fd);VevC=pickle.load(fd);VecY=pickle.load(fd);BCIn=pickle.load(fd);BCIi=pickle.load(fd);Time=pickle.load(fd);TimeOld=pickle.load(fd);\
            TimeEl=pickle.load(fd);TimeNo=pickle.load(fd);TimeS=pickle.load(fd);Step=pickle.load(fd);                Skyline=pickle.load(fd);SDiag=pickle.load(fd);SLen=pickle.load(fd);SymSys=pickle.load(fd);\
            NoLabToNoInd=pickle.load(fd);NoIndToCMInd=pickle.load(fd);ContinuumNodes=pickle.load(fd);CoNoToNoLi=pickle.load(fd);SecDic=pickle.load(fd);LinAlgFlag=pickle.load(fd);ResultTypes=pickle.load(fd);
        fd.close()
#        s = StepList[StepCounter]
        if len(ContinuumNodes)>0: CoorTree = spatial.cKDTree( ContinuumNodes ) # for search purposes, e.g. for EFG or aggregates or embedded truss elements
        else:                     CoorTree = None
    else: 
        raise NameError ("ConFemPostProcStandAlone: .pkl does not exit - correct name ?",FilDir+FilName)
    return NodeList, ElList, MatList, CoorTree, CoNoToNoLi, VecU, NoIndToCMInd, NoLabToNoInd, SecDic, LinAlgFlag, ResultTypes, Time

class ConFemPost:
    def __init__(self):
        pass
    def Run(self, FilDir,fileName,StepCounter, MaPlLib, VTK,VTK3D, Version):
    # load system data
        NodeList,ElemList,MatList,CoorTree,CoNoToNoLi,VecU,NoIndToCMInd,NoLabToNoInd, SecDic, LinAlgFlag, ResultTypes, LastTime = PickleLoad( FilDir, fileName, StepCounter)
        # process plot options
        ScaleDis, PE2DFlag, PE3DFlag = 1.0, True, False
        Post1DFlag, PlotTimes, ScaleStress, PostNodeFlag, ScaleDis2D, ShellL = True, [], [], True, 0., False
        #
        if pth.isfile(fileName+".plt.txt"):                                     # read plot options file if there is any
            f1=open( fileName+".plt.txt", 'r')
            from ConFemInOut import ReadPlotOptionsFile
            ScaleDis, ScaleDis2D, PE2DFlag, PE3DFlag, PlotTimes, Post1DFlag, ScaleStress, PostNodeFlag, ShellL, ScaleShellL = ReadPlotOptionsFile(f1, SecDic) # to modify plot scaling factors
            f1.close()
        # 
        if Version == 1:
            if Post1DFlag:
                f2= open( fileName+".elemout.txt", "r")
                PostElem1D( f2,  ResultTypes, PlotTimes)
                f2.close()
            #
            if PostNodeFlag:
                PostNode( ElemList, NodeList,NoIndToCMInd, VecU, ScaleDis, None)
                print('PostNode   process plot for last time', "%6.4f"%LastTime)
            #
            ff  = open(fileName+".elemout_"+".txt",'r')
            ffN = open(fileName+".nodeout_"+".txt",'r')
            EndFlag, EndFlagN = False, False
            # loop over times in result data sets
            while not EndFlag and not EndFlagN:
                EndFlag, EndFlagN, Time, TimeN, ElemResults, NodeResults = ReadResultsFromFile( ff, ffN)
                if Time != TimeN: raise NameError("ConFemPostProcStandAlone: async time for elements and nodes",Time, TimeN)
                if Time in PlotTimes or len(PlotTimes)==0:
                    if PE2DFlag:
                        SCa, ScaD = PostScales( ElemList,SecDic, ElemResults, ScaleStress)
    #                    _ = PostElem2D( ElemList,NodeList, ScU,[], SecDic, MaPlLib, NoIndToCMInd, ElemResults,NodeResults, Time, FilDir+"/", SCa)
                        _ = PostElem2D( ElemList,NodeList, ScaleDis2D, SecDic, MaPlLib, NoIndToCMInd, ElemResults,NodeResults, Time, FilDir+"/", SCa)
                    if PE3DFlag:
                        print('PostNode3D process plot for time ', Time)
                        PostNode3D( ElemList,NodeList,NoIndToCMInd, NodeResults,[], Time, ScaleDis)
                        l = PostElem3D( ElemList, NodeList,NoIndToCMInd, ElemResults, Time)
                        if l>0: print('PostElem3D process plot for time ', Time)
                    if VTK:
                        PostElemVTK(FilDir,fileName, str(int(1000*round(Time,5))), ElemList, NodeList,NoIndToCMInd, VecU, ElemResults,NodeResults, LinAlgFlag, VTK3D)
            ff.close()
            ffN.close() 
            #
            if ShellL: 
                f2=open( fileName+".elemout.txt", 'r')
                PostElemSH4( ElemList, NodeList,NoIndToCMInd, MatList, f2, SCa, 1., 1., PlotTimes, ScaleShellL )
                f2.close()
            #   
            if pth.isfile(fileName+".opt.txt"): 
                f4=open( fileName+".opt.txt", 'r')
                WrNodes, _, _, _ = ReadOptionsFile(f4, NodeList, NoLabToNoInd, NoIndToCMInd)
                f4.close()
            else: WrNodes = []
            if len(WrNodes)>0:
                f5=open( fileName+".timeout.txt", 'r')
                print('PostHist process plot for history')
                PlotHist( f5, WrNodes )
                f5.close()
            #
            return True
        elif Version == 2:
            LogName="../LogFiles"                                               # to log temporary data
            from ConFemInOut import PlotResiduals
            PlotResiduals( LogName, "log", 20, 1, 1.0e+0)
            return True
        elif Version == 3:
            LogName="../LogFiles"                                               # to log temporary data
            from ConFemInOut import WriteResiduals
            WriteResiduals( LogName, 1, 8, 100)
            return False
            
    ############################################################################################################################### # ts
if __name__ == "__main__":
    StepCounter = 0
    MaPlLib = True                                                          # flag for using matplotlib 
    VTK,VTK3D = False, False
    # ScP scaling factor principal stresses, ScU scaling factor displacements
    #                                                                  [output times], scaling factor displacements, [[Section 1 scaling factors], [Section 2 scaling factors], .... ]
    #                                                                                                                each Section may display more than one result depending on element type
    # DataExamples
#    Name = "../DataExamples/E03/E3-01"  # 01, 04
#    Name = "../DataExamples/E04/E4-06".
#    Name="../DataExamples/E04/E3-02_CircMises"                              # 2_CircMises"
#    Name ="../DataExamples/E05/E5-02"
    Name = "../DataExamples/E07/E7-01a"
#    Name = "../DataExamples/E07/E7-05"                                   # 1D phase field
#    Name = "../DataExamples/E07-06/E7-06a"                               # SDA
#    Name = "../DataExamples/E08-02/ElasticLT/E8-02"                      # deep beam nonlinear
#    Name = "../DataExamples/E08/E8-03"
    Name = "../DataExamples/E08/E8-04"
    Name = "../DataExamples/E09/E9-05c"
    Name = "../DataExamples/E10/E10-01c"
    
#    Name = "../DataExamples/E10-02/IsoDam/E10-02"
    #
    Name = "../_DataShellsSlabs/Shell" # Shell, bridge_el05m, c_1461(0.08)_2.1e5_0.3_segment load
    #
    Name = "C:/Users/uhc/Documents/Work/Edu/VorlesungenUebungen/CE_RC_2022TUM/ConFem/NotchedPlate/E7-01a"
    # 
    print('ConFemPost for ',Name)
    Version = 1
    ComFemPost_ = ConFemPost()
    Flag = ComFemPost_.Run( "",Name,StepCounter, MaPlLib, VTK,VTK3D, Version)
    print('finish processing')
    if MaPlLib or Flag: plt.show() 
    print('finish all')
